package coms;

import java.io.BufferedReader;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.rmi.registry.LocateRegistry;
import java.util.Iterator;
import java.util.Set;
import javax.management.Attribute;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import org.mortbay.jetty.Server;
import org.xml.sax.SAXException;
import coms.config.ConfigOptions;
import coms.config.ConfigParser;

/**
 * This class takes care of starting up the system and shutting it down
 */
public class CoMS {
	private static MBeanServer server;
	private static String componentDomain;
	private ObjectName agent;
	private Server jettyserver;

	/**
	 * Main just calls contructor
	 */
	public static void main(String args[]) {
		new CoMS();
	}

	//TODO Remove me
	/*
	 * Gets the MBeanServer handle for a specific domain
	 * 
	 * @param domain
	 *            The name of the domain we want to get server handle to.
	 *
	private MBeanServer getServerHandle(String domain) {
		MBeanServer serv = null;
		ArrayList mbeanservers = MBeanServerFactory.findMBeanServer(null);
		Iterator iter = mbeanservers.iterator();
		while (iter.hasNext()) {
			serv = (MBeanServer) iter.next();
			if (serv.getDefaultDomain().equals(domain))
				break;
		}
		if (serv == null) {
			System.out.println("Could not find mbeanserver handle");
			System.exit(0);
		}
		return serv;
	}*/

	/**
	 * Gets the MBeanServer handle
	 */
	public static MBeanServer getServerHandle() {
		return server;
	}

	/**
	 * Gets the component domain
	 */
	public static String getComponentDomain() {
		return componentDomain;
	}

	/**
	 * Starts the component loader MBean thread
	 */
	private void loadComponents() throws Exception {
		ObjectName loader = createObjectName("ComponentLoader", "Core");
		server.invoke(loader, "loadComponents", null, null);
	}

	/**
	 * Creates an object name which specifies a module
	 * 
	 * @param name
	 *            The name of the object
	 * @param type
	 *            The type of the object
	 * @return The created ObjectName
	 */
	private ObjectName createObjectName(String name, String type) {
		try {
			return new ObjectName(server.getDefaultDomain() + ":Name=" + name
					+ ", Type=" + type);
		}
		catch (MalformedObjectNameException e) {
			System.out.println("Malformed object name");
			e.printStackTrace(System.out);
			System.exit(3);
		}
		return null;
	}

	/**
	 * Uses the configuration options to create a new MBeanserver with the
	 * module domain
	 * 
	 * @param config
	 *            The configuration options
	 */
	private void createServer(ConfigOptions config) throws Exception {
		// Create a Managed Bean Server
		server = MBeanServerFactory.createMBeanServer(config.getModuleDomain());
	}

	/**
	 * Registers and configures the Logger MBean
	 * 
	 * @param config
	 *            The configuration options
	 */
	private void initLogger(ConfigOptions config) throws Exception {
		// Logger object name
		ObjectName logger = createObjectName(config.getLoggerName(), config
				.getModuleType());
		// Register logger to mbeanserver
		server.createMBean(config.getLoggerClass(), logger);
		// Set logfile
		server.setAttribute(logger, new Attribute("File", config.getLogFile()));
		// Clear the logfile
		if (config.getClrLogAtStart())
			server.invoke(logger, "clear", null, null);
		// Set logger status ON
		server.setAttribute(logger, new Attribute("On", config
				.getLoggerStatus()));
		log("System starting...");
		log("Module loaded: " + logger.toString());
	}

	/**
	 * Registers and configures the ModuleAgent MBean
	 * 
	 * @param config
	 *            The configuration options
	 */
	private void initModuleAgent(ConfigOptions config) throws Exception {
		// Create agent object name
		agent = createObjectName("ModuleAgent", config.getModuleType());
		// Register module agent
		server.createMBean(config.getModuleAgentClass(), agent);
		// Set CoMS agent mlet url
		server.setAttribute(agent, new Attribute("URL", config.getMLetPath()));
		// Set config options in agent
		server.setAttribute(agent, new Attribute("Config", config));
		log("Module loaded: " + agent.toString());
	}

	/**
	 * Uses the ModuleAgent Mbean to load all modules
	 */
	private void loadModules() throws Exception {
		server.invoke(agent, "loadModules", null, null);
	}

	/**
	 * Uses the Logger MBean to record a message
	 * 
	 * @param msg
	 *            The message to be logged
	 */
	private void log(String msg) throws Exception {
		ObjectName logger = createObjectName("Logger", "Core");
		server.invoke(logger, "log", new Object[] { msg },
				new String[] { String.class.getName() });
	}

	/**
	 * Uses the ModuleAgent MBean to register and start the RMI (server) service
	 * 
	 * @param config
	 *            The configuration options
	 */
	private Boolean startRMIService(ConfigOptions config) throws Exception {
		Boolean success = false;
		// Create and register the RMI connector mbean
		server.invoke(agent, "createRMIConnector", new Object[] { config
				.getModuleListenURL() },
				new String[] { String.class.getName() });
		// Start the RMI service
		success = (Boolean) server.invoke(agent, "startRMIServer", null, null);
		return success;
	}

	/**
	 * Uses the ModuleAgent MBean to stop the RMI (server) service
	 */
	private void stopRMIService() throws Exception {
		// Start the RMI service
		server.invoke(agent, "stopRMIServer", null, null);
	}

	/**
	 * Uses the ComponentManager MBean to stop/terminate all components loaded
	 * in the system
	 * 
	 * @param config
	 *            The configuration options
	 */
	private void stopComponents(ConfigOptions config) throws Exception {
		ObjectName comManager = createObjectName("ComponentManager", "Core");
		Set cnames = (Set) server.getAttribute(comManager,
				"ComponentObjectNames");
		Iterator clist = cnames.iterator();
		while (clist.hasNext()) {
			ObjectName component = (ObjectName) clist.next();
			server.invoke(component, "stop", null, null);
		}
		log("All components have been terminated gracefully");
	}

	/**
	 * Get the configuration options from the config file
	 * 
	 * @param configfile
	 *            The configuration file
	 * @return The configuration options
	 */
	private ConfigOptions getConfigOptions(String configfile) {
		try {
			ConfigParser parser = new ConfigParser(configfile);
			return parser.parse();
		}
		catch (SAXException e) {
			e.printStackTrace();
			System.exit(0);
		}
		return null;
	}

	/**
	 * Initializes and start the Jetty webserver
	 * 
	 * @param configfile
	 *            The configuration file
	 */
	private void startJetty(String configfile) {
		try {
			jettyserver = new Server(configfile);
			jettyserver.start();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Stops the Jetty webserver
	 */
	private void stopJetty() {
		try {
			jettyserver.stop();
		}
		catch (Exception e) {
			System.out.println(e.toString());
		}
	}

	/**
	 * Shutdowns all databases
	 */
	private void shutdownDatabases(String[] databases) throws Exception {
		for (int i = 0; i < databases.length; i++) {
			ObjectName db = createObjectName(databases[i], "Core");
			server.invoke(db, "shutdown", null, null);
		}
	}

	/**
	 * Creates a display menu for the CoMS
	 */
	private void menu() {
		char[] c = { (char) 186 };
		String d = "  " + new String(c);
		System.out.print("\n  ");
		System.out.print((char) 201);
		for (int i = 0; i < 38; i++)
			System.out.print((char) 205);
		System.out.println((char) 187);
		System.out.println(d + "                                    " + d);
		System.out.println(d + "           //  CoMS  //             " + d);
		System.out.println(d + "                                    " + d);
		System.out.println(d + "    .::. All modules loaded  .::.   " + d);
		System.out.println(d + "    .::. System running...   .::.   " + d);
		System.out.println(d + "                                    " + d);
		System.out.println(d + "           [1] - Quit               " + d);
		System.out.println(d + "                                    " + d);
		System.out.print("  ");
		System.out.print((char) 200);
		for (int i = 0; i < 38; i++)
			System.out.print((char) 205);
		System.out.println((char) 188);
		System.out.print("  : ");
	}

	/**
	 * Displays a menu until the users presses [1]
	 */
	private void display() {
		String choice;
		BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
		menu();
		try {
			while (true) {
				choice = br.readLine();
				if (choice.equals("1")) {
					System.out.println("CoMS system is exiting...");
					break;
				}
				else {
					menu();
				}
			}
		}
		catch (IOException e) {}
	}

	/**
	 * The CoMS constructor
	 */
	CoMS() {
		try {
			// Get config options
			ConfigOptions config = getConfigOptions("./config/config.xml");
			// Start RMI registry
			LocateRegistry.createRegistry(new Integer(config.getRMIShellArg())
					.intValue());
			// Create new MBean server
			createServer(config);
			// Init component domain
			componentDomain = config.getComponentDomain();
			// Initialize logger
			initLogger(config);
			// Initilize ModuleAgent
			initModuleAgent(config);
			// Load all modules
			loadModules();
			log("All modules are loaded");
			// Load all components
			loadComponents();
			log("All components loaded");
			// Try starting the RMI (server) service
			if (startRMIService(config))
				log("RMI Service is started");
			// Start the jetty webserver
			startJetty("./jetty/deploy_jetty.xml");
			// Display menu
			display();
			// Graceful termination of the component threads
			stopComponents(config);
			// Stop the RMI server
			stopRMIService();
			// Stop the jetty webserver
			stopJetty();
			shutdownDatabases(new String[] { "HSQL", "Oracle", "Derby" });
			// Force exit because we dont want to wait for RMI registry to
			// finish
			System.exit(0);
		}
		catch (MalformedObjectNameException e) {
			System.out.println("Name for component has invalid characters");
			System.exit(1);
		}
		catch (ReflectionException e) {
			System.out.println("Unknown error");
			e.printStackTrace(System.out);
			System.exit(6);
		}
		catch (InstanceAlreadyExistsException e) {
			System.out
					.println("Cannot create component because a component with this name already exists");
			System.exit(2);
		}
		catch (MBeanRegistrationException e) {
			System.out
					.println("Something went wrong with the registration of the component to the agent");
			e.printStackTrace(System.out);
			System.exit(4);
		}
		catch (MBeanException e) {
			System.out
					.println("Something went wrong with registering the component to the agent");
			e.printStackTrace(System.out);
			System.exit(11);
		}
		catch (NotCompliantMBeanException e) {
			System.out
					.println("The component is not compliant according to specification");
			System.exit(12);
		}
		catch (InstanceNotFoundException e) {
			System.out
					.println("The mbean that corresponds to the objectname could not be found");
			e.printStackTrace(System.out);
			System.exit(3);
		}
		catch (Exception e) {
			System.out.println("An unknown exception occured");
			e.printStackTrace(System.out);
			System.exit(3);
		}
	}

	public static void setComponentDomain(String componentDomain) {
		CoMS.componentDomain = componentDomain;
	}
}

/**
 * Class that specifies which files should be filtered when retrieving a file
 * list from a directory
 */
class ExtensionFilter implements FilenameFilter {
	private String extension;

	public ExtensionFilter(String extension) {
		this.extension = extension;
	}

	public boolean accept(File dir, String name) {
		return (name.endsWith(extension));
	}
}
