/*
 * @(#)file      Component.java
 * @(#)author    Mats Mortensen
 * @(#)version   0.9
 * @(#)lastedit  29.4.2005
 * 
 */
package coms.component;

import java.io.File;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import javax.management.Attribute;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanException;
import javax.management.MalformedObjectNameException;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import coms.CoMS;
import coms.core.Module;
import coms.core.Util;
import coms.metadata.MetaData;
import coms.metadata.MetaDataWriter;

/**
 * Class which provides core functionality for user components Functionality
 * includes mbean interface, database interface, thread interface
 */
public abstract class Component extends Module implements ComponentMBean,
		Runnable {
	protected MetaData metadata;
	protected String componentID;
	private Thread componentThread = null;
	private String extDir;
	private String webDir;

	public Component() {
		super();
	}

	/**
	 * Initialize the metadata file describing this component Initialize the
	 * parameters used by the database access layer
	 * 
	 * @param DatabaseHandler
	 *            object, MetaData object
	 */
	public final void initialize(MetaData data) {
		metadata = data;
		componentID = data.getID();
		// Specific initializations for input, processing and output components
		doInit();
		// Specific init by user component
		init();
	}

	/**
	 * Getters and setters for web and ext dirs
	 */
	public String getExtDir() {
		return this.extDir;
	}

	public void setExtDir(String dir) {
		this.extDir = dir;
	}

	public String getWebDir() {
		return this.webDir;
	}

	public void setWebDir(String dir) {
		this.webDir = dir;
	}

	/**
	 * Method that will be overloaded to execute component code when a component
	 * is stopped
	 */
	public abstract void unload();

	/**
	 * Method that will be overloaded to initialize more in the specific
	 * components
	 */
	public abstract void doInit();

	/**
	 * Method that will be overloaded by user components to be able to execute
	 * code
	 */
	public abstract void execute() throws Exception;

	/**
	 * Method that will be overloaded by user component to be able to initialize
	 */
	public abstract void init();

	public abstract void addNotificationListener(NotificationListener listener,
			NotificationFilter filter, Object handback);

	public abstract void removeNotificationListener(
			NotificationListener listener);

	/**
	 * Method that will be overloaded by user components to be able to execute
	 * code when an event happens. (ie. table they are subscribed to gets
	 * changed)
	 */
	public abstract void onEvent(String table);

	/**
	 * Gets the metadata file describing this component
	 * 
	 * @return MetaData object
	 */
	public final MetaData getMetaData() {
		return this.metadata;
	}

	/**
	 * Subscribes a component to a table using the EventHandler.
	 * 
	 * @param String
	 *            table
	 */
	public final void subscribe(String table) {
		try {
			ObjectName eh = createObjectName("EventHandler", "Core");
			server.invoke(eh, "subscribe", new Object[] { metadata.getName(),
					metadata.getType(), metadata.getVersion(), table },
					new String[] { String.class.getName(),
							String.class.getName(), String.class.getName(),
							String.class.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Method available through FileHandler to components
	 */
	public final void setMaxFiles(Integer max) {
		try {
			ObjectName fh = createObjectName("FileHandler", "Core");
			server.setAttribute(fh, new Attribute("maxFiles", max));
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Method available through FileHandler to components
	 */
	public final String putFile(InputStream stream) throws Exception {
		ObjectName fh = createObjectName("FileHandler", "Core");
		return (String) server.invoke(fh, "putFile", new Object[] { stream },
				new String[] { InputStream.class.getName() });
	}

	/**
	 * Method available through FileHandler to components
	 */
	public final Integer getMaxFiles() {
		try {
			ObjectName fh = createObjectName("FileHandler", "Core");
			return (Integer) server.getAttribute(fh, "maxFiles");
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * Method available through FileHandler to components
	 */
	public final String putFile(byte[] data) throws Exception {
		Byte[] bytes = Util.bytesToByteObj(data);
		ObjectName fh = createObjectName("FileHandler", "Core");
		return (String) server.invoke(fh, "putFile", new Object[] { bytes },
				new String[] { Byte[].class.getName() });
	}

	public final void putFile(String fileId, byte[] data) throws Exception {
		Byte[] bytes = Util.bytesToByteObj(data);
		ObjectName fh = createObjectName("FileHandler", "Core");
		server
				.invoke(fh, "putIntoFile", new Object[] { fileId, bytes },
						new String[] { String.class.getName(),
								Byte[].class.getName() });
	}

	/**
	 * Method available through FileHandler to components
	 */
	public final InputStream getFile(String path) throws Exception {
		ObjectName fh = createObjectName("FileHandler", "Core");
		return (InputStream) server.invoke(fh, "readFile",
				new Object[] { path }, new String[] { String.class.getName() });
	}

	/**
	 * Method available through FileHandler to components
	 */
	public final void deleteFile(String path) throws Exception {
		ObjectName fh = createObjectName("FileHandler", "Core");
		server.invoke(fh, "deleteFile", new Object[] { path },
				new String[] { String.class.getName() });
	}

	/**
	 * Stops the component thread. Actually causes it to terminate, but can be
	 * restarted with all variables intact.
	 */
	public final void stop() {
		ObjectName name = null;
		try {
			componentThread = null;
			unload();
			// Update config options from component variables
			HashMap opt = metadata.getCustomOptions();
			HashMap<String, String> newopt = new HashMap<String, String>();
			name = createObjectName(CoMS.getComponentDomain(), metadata
					.getName(), metadata.getType());
			if (opt != null) {
				if (!opt.isEmpty()) {
					Set set = opt.keySet();
					Iterator it = set.iterator();
					while (it.hasNext()) {
						String key = (String) it.next();
						key = key.substring(0, 1).toUpperCase()
								+ key.substring(1);
						String value = (String) server.getAttribute(name, key);
						newopt.put(key, value);
					}
					opt = newopt;
				}
			}
			String serviceName = metadata.getService();
			File dir = new File("deploy/metapersistence/" + serviceName + "/");
			// Save config options
			MetaDataWriter.write(metadata, dir);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * The run method is part of the runnable interface. It is the main method
	 * for running threads. When this method terminates the thread "dies".
	 */
	public final void run() {
		Thread thisThread = Thread.currentThread();
		while ((componentThread == thisThread) && (componentThread != null)) {
			try {
				execute();
			}
			catch (Exception e) {
				String msg = "<span style=\"color:red\">Error: </span>An exception occured in "
						+ metadata.getName();
				msg += ". This component will therefore be terminated<br>Caused by: "
						+ e.toString();
				log(msg);
				stop();
				e.printStackTrace();
			}
			try {
				int freq;
				if (metadata.getCustomOptions() != null) {
					if (metadata.getCustomOptions().containsKey(
							"updatefrequency")) {
						String value = (String) metadata.getCustomOptions()
								.get("updatefrequency");
						if (value != null)
							freq = Integer.valueOf(value);
						else
							freq = 2500;
					}
					else {
						freq = 2500;
					}
				}
				else {
					freq = 2500;
				}
				if (freq < 0)
					freq = 2500;
				if (freq < 50)
					freq = 50;
				Thread.sleep(freq);
			}
			catch (InterruptedException e) {
				// Continue working
			}
		}
	}
	
	protected final Object invoke(String mbean, String method, Object... params) 
			throws 	InstanceNotFoundException, MalformedObjectNameException, 
					MBeanException, ReflectionException {
		String[] signature = new String[params.length];
		for (int i = 0; i < params.length; i++) {
			signature[i] = params[i].getClass().getName();
		}
		return server.invoke(new ObjectName(mbean), method, params, signature);
	}
	
	protected final void changeAttributeValue(String mbean, String attributeName, Object newValue) 
			throws 	InstanceNotFoundException, MalformedObjectNameException, 
			MBeanException, ReflectionException, AttributeNotFoundException,
			InvalidAttributeValueException {
		server.setAttribute(new ObjectName(mbean), new Attribute(attributeName, newValue));
	}

	/**
	 * The start method is part of the runnable interface. It initializes the
	 * thread to be run. This method will only be run once when the user
	 * component starts. It can also be used to start a component which has been
	 * stopped.
	 */
	public final void start() {
		if (componentThread == null) {
			componentThread = new Thread(this, metadata.getName());
			// Mark components as Daemon threads so they will terminate when
			// container terminates
			componentThread.setDaemon(true);
			componentThread.start();
		}
	}
}
