/*
 * @(#)file      InputComponent.java
 * @(#)author    Mats Mortensen
 * @(#)version   0.9
 * @(#)lastedit  29.6.2005
 * 
 */
package coms.component;

import java.io.DataInputStream;
import java.io.IOException;
import java.net.ConnectException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.HashMap;
import javax.management.ObjectName;
import coms.tools.inference.Engine;

/**
 * Class which provides core functionality for input components
 */
public abstract class InputComponent extends AbstractNotifier {
	private Binding binding;

	public InputComponent() {
		super();
	}

	/**
	 * Initializes all input components
	 */
	public final void doInit() {
		// Create a new binding type
		this.binding = locateBinding(metadata.getBindingType().toLowerCase(),
				metadata.getCustomOptions());
	}

	/**
	 * Creates a new binding object to be used depending on the binding type the
	 * component has specified
	 * 
	 * @param String
	 *            type
	 */
	private Binding locateBinding(String type, HashMap<String, String> param) {
		Binding binding = null;
		if (type.equals("tcp"))
			binding = new TCPBinding(param.get("host"), new Integer(param
					.get("timeout")).intValue(), new Integer(param.get("port"))
					.intValue());
		else if (type.equals("udp")) {
			binding = new UDPBinding();
		}
		else if (type.equals("xmlrpc")) {
			createXmlRpcBinding();
		}
		return binding;
	}

	/**
	 * Registers an XML-RPC binding
	 */
	private void createXmlRpcBinding() {
		try {
			ObjectName handler = createObjectName("XmlRpcHandler", "Core");
			server.invoke(handler, "addHandler", new Object[] {
					metadata.getName(), this }, new String[] {
					String.class.getName(), Object.class.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	public final Engine createEngine(Boolean isPersistent) {
		try {
			ObjectName factory = createObjectName("InferenceFactory", "Core");
			return (Engine) server.invoke(factory, "createEngine",
					new Object[] { isPersistent, componentID }, new String[] {
							Boolean.class.getName(), String.class.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Reads from sensor
	 */
	public DataInputStream read() throws ClassNotFoundException,
			ConnectException, IOException, SocketTimeoutException {
		return binding.read();
	}

	/**
	 * Writes to sensor
	 */
	public void write(String msg) throws ConnectException,
			UnknownHostException, IOException, SocketTimeoutException {
		binding.write(msg);
	}

	public void write(Object obj) throws ConnectException,
			UnknownHostException, IOException, SocketTimeoutException {
		binding.write(obj);
	}

	public void write(String msg, String host, int port, int timeout)
			throws ConnectException, UnknownHostException, IOException,
			SocketTimeoutException, Exception {
		binding.connect(host, port, timeout);
		binding.write(msg);
	}

	public void write(Object obj, String host, int port, int timeout)
			throws ConnectException, UnknownHostException, IOException,
			SocketTimeoutException, Exception {
		binding.connect(host, port, timeout);
		binding.write(obj);
	}

	public DataInputStream read(String host, int port, int timeout)
			throws ClassNotFoundException, ConnectException, IOException,
			SocketTimeoutException {
		binding.connect(host, port, timeout);
		return binding.read();
	}

	/**
	 * This method is not used by input components
	 */
	public final void onEvent(String table) {
		Thread.yield();
	}
}
