/*
 * @(#)file      JarClassLoader.java
 * @(#)author    Mats Mortensen
 * @(#)version   0.9
 * @(#)lastedit  29.4.2005
 * 
 * @(#)Credit	Based on the JarClassLoader from (C) 2004 Ibermatica S.A.
 * 				@author Alex Fernandez.
 *
 */
package coms.core;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;

/**
 * Loads classes from a .jar file. When initialized, tries to instantiate all
 * classes in the .jar.
 */
public class JarClassLoader extends ClassLoader {
	private HashMap<String, java.lang.Class> classesDefined = new HashMap<String, java.lang.Class>();
	// private HashMap componentsFound = new HashMap();
	private JarLoader jarLoader;
	private boolean redundant;
	public final static char PACKAGE_SEPARATOR = '.';
	public final static char JAR_SEPARATOR = '/';

	/**
	 * Creates a class loader to load all classes within a jar file.
	 * 
	 * @param fileWithJar
	 *            contains the classes to define.
	 */
	public JarClassLoader(File fileWithJar) {
		this.jarLoader = new JarLoader(fileWithJar);
	}

	/**
	 * Loads all classes in a jar file.
	 */
	public void loadAllClasses() throws JarLoaderException {
		this.jarLoader.open();
		this.jarLoader.readAll();
		this.jarLoader.close();
		this.defineAllClasses();
	}

	/**
	 * Try to find the class in the .jar file. Keeps a cache of defined classes,
	 * so that classes already found are returned.
	 */
	// TODO fix the type of the class and remove SupressWarnings
	@SuppressWarnings("unchecked")
	public Class findClass(String name) throws ClassNotFoundException {
		try {
			String validName = convertToClassName(name);
			// all classes are loaded at startup
			if (this.classesDefined.containsKey(validName)) {
				// therefore, if not defined, it's not here
				return (Class) this.classesDefined.get(validName);
			}
			if (this.jarLoader.hasClassDataFor(validName)) {
				return this.defineLoadedClass(validName);
			}
			return super.findClass(name);
		}
		catch (ClassNotFoundException e) {
			String msg = "Class was not found in jar file. <br>Caused by: "
					+ e.getMessage();
			throw new ClassNotFoundException(msg);
		}
	}

	/**
	 * Define all classes found in the jar file.
	 */
	public void defineAllClasses() throws JarLoaderException {
		Iterator iterator = this.jarLoader.classIterator();
		// check if the class loader is redundant,
		// i.e. all classes are already defined
		// just check the first
		if (!iterator.hasNext()) {
			// no classes
			return;
		}
		String name = (String) iterator.next();
		try {
			// try in system class loader first
			this.findSystemClass(name);
			this.redundant = true;
		}
		catch (ClassNotFoundException e) {
			// not defined; not redundant
			this.redundant = false;
		}
		// start again
		iterator = this.jarLoader.classIterator();
		while (iterator.hasNext()) {
			name = (String) iterator.next();
			try {
				this.defineLoadedClass(convertToClassName(name));
			}
			catch (ClassNotFoundException e) {
				String msg = "Class " + name + " not found in " + this
						+ "<br>Caused by: " + e.getMessage();
				throw new JarLoaderException(msg);
			}
		}
	}

	/**
	 * Once we know that the jar loader has data for a class, this method is
	 * called to have the class loaded and defined.
	 * 
	 * @param name
	 *            of the class.
	 * @return Class defined class.
	 */
	private Class defineLoadedClass(String name) throws ClassNotFoundException {
		if (this.classesDefined.containsKey(name)) {
			return (Class) this.classesDefined.get(name);
		}
		byte[] classData = this.jarLoader.getClassData(name);
		Class resultClass;
		if (this.redundant) {
			resultClass = this.findSystemClass(name);
		}
		else {
			resultClass = this
					.defineClass(null, classData, 0, classData.length);
		}
		this.classesDefined.put(name, resultClass);
		return resultClass;
	}

	/**
	 * Converts a class name with dots to class name with slashes.
	 * 
	 * @param original
	 *            the original name of the zip entry containing the class, whose
	 *            name will be in the form "package/package/ClassName"; or the
	 *            class name with slashes, like "package/package.ClassName".
	 * @return the class name in the form "package/package/ClassName".
	 */
	public static String convertToClassName(String original) {
		String withoutExtension = original;
		String className = withoutExtension.replace(PACKAGE_SEPARATOR,
				JAR_SEPARATOR);
		return className;
	}
}
