/*
 * @(#)file      JarLoader.java
 * @(#)author    Mats Mortensen
 * @(#)version   0.9
 * @(#)lastedit  29.4.2005
 * 
 * @(#)Credit	Based on the JarLoader from (C) 2004 Ibermatica S.A.
 * 				@author Alex Fernandez.
 *
 */
package coms.core;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

/**
 * Used to load all class files from a jar file.
 */
public class JarLoader {
	private JarFile jarFile;
	private File fileWithJar;
	private HashMap<String, byte[]> classesFound = new HashMap<String, byte[]>();

	/**
	 * Construct a jar loader for the given file.
	 * 
	 * @param fileWithJar
	 */
	public JarLoader(File fileWithJar) {
		this.fileWithJar = fileWithJar;
	}

	/**
	 * Returns filename of jar file
	 */
	public String getFileName() {
		return this.fileWithJar.getName();
	}

	/**
	 * Create the jar file and open it.
	 * 
	 * @throws JarLoaderException
	 *             if there is an I/O exception.
	 */
	public void open() throws JarLoaderException {
		try {
			this.jarFile = new JarFile(this.fileWithJar);
		}
		catch (IOException e) {
			String msg = "Invalid jar file in "
					+ this.fileWithJar.getAbsolutePath();
			throw new JarLoaderException(msg);
		}
	}

	/**
	 * Read all .class entries from the jar file
	 * 
	 * @throws JarLoaderException
	 *             if any of them fail or if no classes where in the jar file
	 */
	public void readAll() throws JarLoaderException {
		boolean hasClassFile = false;
		Enumeration entries = this.jarFile.entries();
		while (entries.hasMoreElements()) {
			ZipEntry zipEntry = (ZipEntry) entries.nextElement();
			if (!zipEntry.isDirectory()
					&& zipEntry.getName().endsWith(".class")) {
				hasClassFile = true;
				this.readClass(zipEntry);
			}
		}
		if (!hasClassFile) {
			try {
				this.jarFile.close();
			}
			catch (IOException e) {}
			String msg = "Jar file had no class files";
			throw new JarLoaderException(msg);
		}
	}

	/**
	 * Find out if the jar file contains the definition of a class.
	 * 
	 * @param name
	 *            of the desired class.
	 * @return boolean if the jar file contained the class when it was read.
	 */
	public boolean hasClassDataFor(String name) {
		return this.classesFound.containsKey(name);
	}

	/**
	 * Check if a file exists in the jar file
	 * 
	 * @return boolean
	 */
	public boolean fileExists(String filename) {
		ZipEntry entry = this.jarFile.getEntry(filename);
		if (entry == null)
			return false;
		else
			return true;
	}

	public void copyFolder(String src, String dest) throws JarLoaderException {
		try {
			File destFile;
			byte[] buf = new byte[1024];
			int bytesread = buf.length;
			File destDir = new File(dest);
			if (!destDir.exists())
				destDir.mkdir();
			Enumeration entries = this.jarFile.entries();
			while (entries.hasMoreElements()) {
				ZipEntry entry = (ZipEntry) entries.nextElement();
				if (entry.getName().substring(0, src.length()).equals(src)) {
					if (!entry.isDirectory()) {
						String filename = entry.getName();
						int index = filename.lastIndexOf("/");
						filename = filename.substring(index + 1);
						destFile = new File(destDir, filename);
						destFile.createNewFile();
						DataInputStream instream = new DataInputStream(
								this.jarFile.getInputStream(entry));
						FileOutputStream outstream = new FileOutputStream(
								destFile);
						while (instream.available() != 0) {
							bytesread = instream.read(buf, 0, buf.length);
							outstream.write(buf, 0, bytesread);
						}
						instream.close();
						outstream.close();
					}
				}
			}
		}
		catch (IOException e) {
			throw new JarLoaderException(
					"Could not get jarfile inputstream. <br>Caused by: IOException: "
							+ e.getMessage());
		}
	}

	/**
	 * Read Metafile.xml from the jar file
	 * 
	 * @return inputstream of the metafile
	 * @throws JarLoaderException
	 *             if no metafile is found or if io exception
	 */
	public InputStream readXMLFile(String filename) throws JarLoaderException,
			JarValidityException {
		ZipEntry entry = this.jarFile.getEntry(filename);
		if (entry == null) {
			String msg = "The file \"" + filename + "\" in "
					+ this.jarFile.getName() + " can not be found";
			throw new JarValidityException(msg);
		}
		try {
			InputStream XMLfile = this.jarFile.getInputStream(entry);
			return XMLfile;
		}
		catch (IOException e) {
			throw new JarLoaderException(
					"Could not get jarfile inputstream. <br>Caused by: IOException: "
							+ e.getMessage());
		}
	}

	/**
	 * Close the jar file.
	 * 
	 * @throws JarLoaderException
	 *             if there is an I/O exception.
	 */
	public void close() throws JarLoaderException {
		try {
			this.jarFile.close();
		}
		catch (IOException e) {
			String msg = "Could not close jar file " + this.jarFile.getName()
					+ "<br>Caused by: " + e.getMessage();
			throw new JarLoaderException(msg);
		}
	}

	/**
	 * Return an iterator over all classes in the jar file.
	 * 
	 * @return Iterator over classes found in the jar file.
	 */
	public Iterator classIterator() {
		return this.classesFound.keySet().iterator();
	}

	public HashMap<String, byte[]> getClasses() {
		return this.classesFound;
	}

	/**
	 * Get the class definition as a byte array for the class.
	 * 
	 * @param name
	 *            of the desired class.
	 * @return byte[] definition of the class.
	 */
	public byte[] getClassData(String name) {
		return (byte[]) this.classesFound.get(name);
	}

	/**
	 * Uses helper function to read class data into a byte array Adds the class
	 * to the list of found classes
	 * 
	 * @param zipentry
	 *            of the desired class.
	 */
	private void readClass(ZipEntry zipEntry) throws JarLoaderException {
		byte[] bytes = this.readClassData(zipEntry);
		String className = extractExtension(zipEntry);
		this.classesFound.put(className, bytes);
	}

	/**
	 * Reads the class data from the class file into a byte array.
	 * 
	 * @param zipentry
	 *            of the desired class.
	 * @return byte[] definition of the class.
	 * @throws JarLoaderException
	 *             if there is an I/O exception.
	 */
	private byte[] readClassData(ZipEntry zipEntry) throws JarLoaderException {
		try {
			int actuallyRead = 0;
			InputStream inStream = this.jarFile.getInputStream(zipEntry);
			byte[] classData = new byte[(int) zipEntry.getSize()];
			int accumulated = 0;
			while (accumulated < classData.length) {
				actuallyRead = inStream.read(classData, accumulated,
						classData.length - accumulated);
				accumulated += actuallyRead;
			}
			return classData;
		}
		catch (IOException e) {
			String msg = "Could not load " + zipEntry.getName() + " from "
					+ this.fileWithJar.getName();
			msg += "Caused by: " + e.getMessage();
			throw new JarLoaderException(msg);
		}
	}

	/**
	 * Takes out the ".class" extension that zip entries have.
	 * 
	 * @param entry
	 *            the zip entry whose name will be transformed.
	 * @return String the zip entry name without the extension.
	 */
	public static String extractExtension(ZipEntry entry) {
		String withoutExtension = entry.getName();
		int extension = withoutExtension.indexOf(".class");
		// Replace slashes with dots
		withoutExtension = withoutExtension.replace('/', '.');
		return withoutExtension.substring(0, extension);
	}
}
