package coms.core.inference;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.drools.*;
import org.drools.io.*;
import java.io.*;
import java.util.HashSet;
import java.util.Iterator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.sql.SQLException;
import coms.database.ConnectionException;
import coms.tools.relational.*;
import coms.tools.inference.Fact;

public class Engine implements coms.tools.inference.Engine{
	
	String rulefile;
	RuleSet ruleset;
	RuleBase ruleBase;
	WorkingMemory wm;
	boolean isPersistent;
	String dbtype;
	int nextInsertId = 0;
	String componentID;
	
	Boolean mutex = false;
	
    public Engine(boolean persistent, String componentID, String dbtype)
    {
    	try{
    		this.dbtype = dbtype;
    		this.componentID = componentID;
			this.isPersistent = persistent;
			
			// Run select on db to get last insert id
			Select sel = new Select("InferenceEngine", dbtype);
			sel.column("insertID", Select.MAX);
			List list = sel.execute();
			
			if(list != null){
				Iterator it = list.iterator();				
				Map map = (Map) it.next();

				Integer iid = (Integer) map.get("default");
				
				if(iid != null)
					nextInsertId = iid.intValue();
			}
				
				
		}catch(Exception e){
			e.printStackTrace();
		}
    }
    
    private void checkFactInterface() throws FactException, ClassNotFoundException{
    	HashMap<Integer, Rule> rules = ruleset.getRules();
    	
		// Iterate over rules
		Iterator it = rules.keySet().iterator();
		while(it.hasNext()){
			Integer key = (Integer) it.next();
			Rule rule = (Rule) rules.get(key);
			
			// Iterate over selects
			HashMap<String, String> selects = rule.getSelects();
			Iterator iter = selects.keySet().iterator();
			
			while(iter.hasNext()){
				String id = (String) iter.next();
				String classname = (String) selects.get(id);
				
				// Get interfaces for class
				Class cls = Class.forName(classname);
				Class[] interfaces = cls.getInterfaces();
				
				// Check if it implements fact interface
				boolean found = false;
				for(int i=0;i<interfaces.length;i++){
					if(interfaces[i].getName().equals(Fact.class.getName())){
						found = true;
						break;
					}
				}
				
				if(found == false)
					throw new FactException("The class: "+cls.getName()+" does not implement the Fact interface");
				
			}
		}
    }
	
	public void loadRules(String rulefile) throws SAXException, IOException, IntegrationException, FactException, LoadException{
		try{
		
			// Save rulefile path
			this.rulefile = rulefile;
			
			// Parse rule file
			RuleParser parser = new RuleParser();
			this.ruleset = parser.parse(new File(rulefile));
			
			// Check if classes implement Fact interface
			checkFactInterface();
			
			// Build compatible drools xml 
			String xmlstr = RuleLoader.buildDroolsXml(ruleset);
			
			StringReader reader = new StringReader(xmlstr);
			ruleBase = RuleBaseLoader.loadFromReader(reader);
			wm = ruleBase.newWorkingMemory();
			
			
		}catch(ClassNotFoundException e){
			throw new LoadException("A class could not be found while loading the rulefile: "+rulefile+"\nCaused by: "+e.toString());
		}
		
	}
	
	public void reloadRules() throws SAXException, IOException, IntegrationException, FactException, ClassNotFoundException{
		
		// Check if classes implement Fact interface
		checkFactInterface();
		
		// Create valid drools code
		String xmlstr = RuleLoader.buildDroolsXml(ruleset);
		
		// Create new rule base
		StringReader reader = new StringReader(xmlstr);
		ruleBase = RuleBaseLoader.loadFromReader(reader);
		
		// Create new working memory and copy old working memory
		WorkingMemory wmNew = ruleBase.newWorkingMemory();
		List list = wm.getObjects();
		
		Iterator it = list.iterator();
		while(it.hasNext())
			wmNew.assertObject(it.next());
			
		wm = wmNew;
		
		// Write new rule file
   		BufferedWriter out = new BufferedWriter(new FileWriter(rulefile));
        out.write(xmlstr);
        out.close();
		
	}
	
	public RuleSet getRuleSet(){
		return this.ruleset;
	}
	
	public List getObjects() throws FactException, NoRuleSetLoadedException{
		synchronized(mutex){
			if(wm != null)
				return wm.getObjects();
			else
				throw new NoRuleSetLoadedException("Objects cannot be retrieved because no rulesets are loaded");
		}
	}
	
	public FactHandle assertObject(Object obj) throws FactException, SQLException, ConnectionException, NoRuleSetLoadedException{
		
		// If persistent Engine, store fact in db
		if(isPersistent)	store(obj);
		
		// Increment insertid
		nextInsertId++;
		synchronized(mutex){
			if(wm != null)
				return wm.assertObject(obj);
			else
				throw new NoRuleSetLoadedException("Fact cannot be asserted because no rulesets are loaded");
		}
	}
	
	public void modifyObject(FactHandle handle, Object obj) throws FactException, NoRuleSetLoadedException{
		synchronized(mutex){
			if(wm != null)
				wm.modifyObject(handle, obj);
			else
				throw new NoRuleSetLoadedException("Fact cannot be modified because no rulesets are loaded");
		}
	}
	
	public FactHandle getFactHandle(Object obj) throws FactException, NoRuleSetLoadedException{
		synchronized(mutex){
			if(wm != null)
				return wm.getFactHandle(obj);
			else
				throw new NoRuleSetLoadedException("Could not get fact handle because no rulesets are loaded");
		}
	}
	
	public Object getObject(FactHandle handle) throws FactException, NoRuleSetLoadedException{
		synchronized(mutex){
			if(wm != null)
				return wm.getObject(handle);
			else
				throw new NoRuleSetLoadedException("Object cannot be returned because no rulesets are loaded");
		}
	}
	
	public void retractObject(FactHandle handle) throws FactException, NoRuleSetLoadedException{
		synchronized(mutex){
			if(wm != null)
				wm.retractObject(handle);
			else
				throw new NoRuleSetLoadedException("Object cannot be retracted because no rulesets are loaded");
		}
	}
	
	private void store(Object obj) throws SQLException, ConnectionException{
		Insert insert = new Insert("InferenceEngine", dbtype);
		insert.column("insertID", nextInsertId);
		insert.column("componentID", componentID);
		insert.column("Fact", obj);
		insert.execute();
	}
	
	public void fireRules() throws FactException, NoRuleSetLoadedException{
		synchronized(mutex){
			if(wm != null)
				wm.fireAllRules();
			else
				throw new NoRuleSetLoadedException("Rules cannot be fired because no rulesets are loaded");
		}
	}
    
}



