package coms.core.inference;


import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Schema;
import javax.xml.XMLConstants;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import java.io.IOException;
import java.io.File;
import java.util.HashSet;

public class RuleParser extends DefaultHandler 
{
    private final String COMS_NAMESPACE_URI  	= "http://drools.org/coms";
    private final String JAXP_SCHEMA_LANGUAGE 	= "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
    private final String COMS_SCHEMA_URI		= "coms.xsd";
    
	private SAXParser					parser;
	private File						xmlfile;
	private String						starttag, endtag;
	private String						fullvalue = "";
	
	private RuleSet						ruleset; 
	private Rule						rule;
	private	String						namespace;
	private String						schema;
	private String						selectID;
	private Actions						actions;
	private String						actionContent = "";
	
	// Takes appropriate action when we have a starttag of some type
	public void startElement(String namespaceURI, String localName, String qName, Attributes atts)
		throws SAXException{
   	
		starttag = localName.toLowerCase();
		
		if(starttag.equals("rule-set")){
			String name = atts.getValue("name");
			ruleset.setName(name);
    			
		}else if(starttag.equals("rule")){
			
			// Get name and order attributes
			String name = atts.getValue("name");
			String order = atts.getValue("order");
			String loop = atts.getValue("no-loop");
			
			if(order == null) order = "0";
			if(loop == null) loop = "true";
			
			// Create new rule
			rule = new Rule(name);
			rule.setOrder(order);
			rule.setLoop(loop);
			
		}else if(starttag.equals("select")){
			
			// Get id attribute
			selectID = atts.getValue("id");
			
		}else if(starttag.equals("actions")){
			
			// New Actions object
			actions = new Actions();
		}
	}

	// Takes appropriate action when we have a endtag of come type
	public void endElement(String namespaceURI, String localName, String qName) throws SAXException{

		endtag = localName.toLowerCase();
		starttag = "";
		
		 if(endtag.equals("rule")){
			
			// Add current rule to rule list
			ruleset.add(rule);
			
		}else if(endtag.equals("select")){
			
			// Add new select to rule
			rule.addSelect(selectID, fullvalue);
			
		}else if(endtag.equals("condition")){
			
			// Create new condition object
			Condition cond = new Condition(fullvalue);	
		
			// Add new condition to rule
			rule.addCondition(cond);	
		
		}else if(endtag.equals("actions")){
			
			// Set java actions
			actions.setContent(actionContent);
			actionContent = "";
			
			// Set rule actions
			rule.setActions(actions);
			
		}else if(endtag.equals("insert")){
			
			// Add insert to actions
			actions.addInsert(fullvalue);
		
		}else if(endtag.equals("update")){
			
			// Add update to actions
			actions.addUpdate(fullvalue);
			
		}else if(endtag.equals("remove")){
			
			// Add remove to actions
			actions.addRemove(fullvalue);	
		}
			
		
		

		starttag = "";
		fullvalue = new String("");
		endtag = "";
	}

	// Gets the content between the start and end tag
	public void characters(char[] text, int start, int length)throws SAXException 
	{
		String value = new String(text, start, length);		
		value = value.trim();
		fullvalue += value;
		
		// Save action content
		if(starttag.equals("actions")){
			actionContent += fullvalue;	
			fullvalue = "";
		}
	}


	public RuleParser() throws SAXException
	{
		try 
		{
			ruleset = new RuleSet();
			parser = null;
			
            SAXParserFactory factory = SAXParserFactory.newInstance( );
            factory.setNamespaceAware( true );
            factory.setValidating( true );
            
            try
            {
            	parser = factory.newSAXParser( );
            }
            catch ( ParserConfigurationException e )
            {
            	throw new RuntimeException( e.getMessage( ) );
           	}
           	
			try
            {
                parser.setProperty( JAXP_SCHEMA_LANGUAGE,
                                         XMLConstants.W3C_XML_SCHEMA_NS_URI );
                                         
				parser.setProperty("http://java.sun.com/xml/jaxp/properties/schemaSource", new File("config/coms.xsd"));
            
				 
            }
            catch ( SAXNotRecognizedException e )
            {
                System.err.println( "Your SAX parser is not JAXP 1.2 compliant - turning off validation." );
                parser = null;
            }
            
            if ( parser == null )
            {
                // not jaxp1.2 compliant so turn off validation
                try
                {
                    factory.setValidating( false );
                    parser = factory.newSAXParser( );
                }
                catch ( ParserConfigurationException e )
                {
                    throw new RuntimeException( e.getMessage( ) );
                }
            }
            
        	if ( !parser.isNamespaceAware( ) )
        	{
        	    throw new RuntimeException( "parser must be namespace-aware" );
        	}
    	
		}
		catch (SAXNotRecognizedException ex)
		{
			System.out.println("Installed XML parser cannot validate");
			System.exit(0);
		}
		catch (SAXNotSupportedException e)
		{
			System.out.println("Installed XML parser cannot validate");
			System.exit(0);
		}
		catch(Exception e){
			e.printStackTrace();
		}
	}
	
	// Initiate parsing of xml document
	public RuleSet parse(File file) throws SAXParseException, SAXException, IOException
	{		
	
		// Set xml file
		this.xmlfile = file;
	
		// Parse the xml file
		parser.parse(xmlfile, this);
		
		// Return rule list
		return ruleset;
	}
	
	public void error(SAXParseException e) throws SAXException{
		String msg = "<span style=\"color:red\">Error: </span> The metadata file \""+xmlfile.getName()+"\" did not validate. <br> Caused by: "+e.getMessage();
		throw new SAXException(msg);
	}
	
	public void fatalError(SAXParseException e) throws SAXException{
		String msg = "<span style=\"color:red\">Fatal error: </span> The metadata file \""+xmlfile.getName()+"\" did not validate. <br> Caused by: "+e.getMessage();
		throw new SAXException(msg);
	}
	
	public void warning(SAXParseException e) throws SAXException{
		String msg = "<span style=\"color:red\">Warning: </span> The metadata file \""+xmlfile.getName()+"\" did not validate. <br> Caused by: "+e.getMessage();
		throw new SAXException(msg);
	}

	public void setDocumentLocator(Locator locator) {}
	public void startDocument() throws SAXException {}
	public void endDocument() throws SAXException {}
	public void ignorableWhitespace(char[] text, int start, int length) throws SAXException {}
	public void processingInstruction(String target, String data) throws SAXException {}
	public void startPrefixMapping(String prefix, String uri) throws SAXException {}
	public void endPrefixMapping(String prefix) throws SAXException {}
	public void skippedEntity(String name) throws SAXException {}
}