package coms.database;

import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import coms.core.Module;
import coms.metadata.Table;

/**
 * Gives out access to the Derby database in the form of DerbyConnections. Will
 * also monitor the Derby database
 * 
 * @author Tom Arild Jakobsen & Mats Mortensen
 */
public abstract class Database extends Module {
	protected Set<DatabaseConnection> activeCon = Collections
			.synchronizedSet(new HashSet<DatabaseConnection>());
	protected Set<DatabaseConnection> unusedCon = Collections
			.synchronizedSet(new HashSet<DatabaseConnection>());
//	TODO Remove me
	//private HashMap<String, Driver> driverMap = new HashMap<String, Driver>();
	private String protocol;
	private Properties props;

	/**
	 * Abstract methods which each database MBean needs to implement
	 */
	protected abstract String getCreateTableSql(Table tbl) throws SQLException,
			ConnectionException;

	protected abstract String toSqlType(String type);

	public abstract Table getTableInfo(String table) throws SQLException,
			ConnectionException;

	/**
	 * Initiates database variables
	 */
	protected void init(String driver, String protocol, Properties props) {
		try {
			Class.forName(driver);
			this.protocol = protocol;
			this.props = props;
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Takes a slot in the connection pool, creates a connection and returns it
	 * 
	 * @return a connection
	 * @throws ConnectionException
	 *             if there are no free connections
	 */
	public final DatabaseConnection getConnection() throws ConnectionException {
		// Try to recycle an old connection
		synchronized (unusedCon) {
			if (!unusedCon.isEmpty()) {
				// Gets the first element in the set
				DatabaseConnection con = unusedCon
						.toArray(new DatabaseConnection[0])[0];
				unusedCon.remove(con);
				activeCon.add(con);
				return con;
			}
			// Create a new connection
			else {
				DatabaseConnection con = new DatabaseConnection(protocol,
						props, activeCon, unusedCon);
				activeCon.add(con);
				return con;
			}
		}
	}

	/**
	 * Terminates connection to database and removes the connections entry in
	 * the connection pool
	 * 
	 * @param con
	 *            Connection to be terminated
	 */
	public final void terminateConnection(DatabaseConnection con) {
		// Tries to remove it from both active and unused connections
		con.close();
		activeCon.remove(con);
		unusedCon.remove(con);
	}

	/**
	 * Shut downs the database and closes all connection pools
	 */
	public final void shutdownDB(String sqlStr) throws SQLException,
			ConnectionException {
		// Wait until all active connections are unactive
		if (activeCon.size() > 0) {
			try {
				System.out.println("Connections to db still in use...sleeping");
				Thread.sleep(5);
			}
			catch (InterruptedException e) {}
		}
		// Close all unused connections
		Iterator it = unusedCon.iterator();
		while (it.hasNext()) {
			DatabaseConnection con = (DatabaseConnection) it.next();
			con.close();
		}
		// Close any remaining active connections
		it = activeCon.iterator();
		while (it.hasNext()) {
			DatabaseConnection con = (DatabaseConnection) it.next();
			con.close();
		}
		// Execute any final commands
		if (sqlStr != null) {
			DatabaseConnection con = new DatabaseConnection(protocol, props,
					activeCon, unusedCon);
			con.executeSQL(sqlStr, false);
			con.close();
		}
	}

	/**
	 * Connections attribute getter
	 * 
	 * @return total number of connections handeled
	 */
	public final Integer getConnections() {
		return activeCon.size();
	}

	/**
	 * Uses the database to create a Table object to the corresponding name
	 * 
	 * @param name
	 *            the name of the table to get info about, SQL string to select
	 *            the column names from the table, database specific name of
	 *            "column name", database specific name of "column datatype",
	 *            database specific name of "column default"
	 * @return the complete metainfo of the table
	 */
	protected final Table getTableDesc(String tablename, String sqlStr,
			String colName, String colDataType, String colDefault)
			throws SQLException, ConnectionException {
		boolean exists = false;
		// Creates a new table
		Table table = new Table(tablename);
		DatabaseConnection con = new DatabaseConnection(protocol, props,
				activeCon, unusedCon);
		// Get table metadata
		DatabaseMetaData meta = con.getMetaData();
		ResultSet rs = meta.getColumns(null, null, tablename.toUpperCase(),
				null);
		// Iterate over columns
		while (rs.next()) {
			// Table exists
			exists = true;
			String colname = rs.getString("COLUMN_NAME");
			String coltype = rs.getString("DATA_TYPE");
			String coldefault = rs.getString("COLUMN_DEF");
			if (coldefault == null)
				coldefault = new String("");
			table.addField(colname, coltype, coldefault);
		}
		if (exists == false)
			return null;
		return table;
	}

	/**
	 * Adds a table to the database
	 * 
	 * @param tbl
	 *            The table to be added to the database
	 * @return wheater the table was added successfully
	 */
	public final Boolean createTable(Table tbl) throws SQLException,
			ConnectionException {
		DatabaseConnection con = new DatabaseConnection(protocol, props,
				activeCon, unusedCon);
		// Create primary ID field if not present
		if (!tbl.hasPrimaryKey()) {
			if (!tbl.hasField("ID")) {
				tbl.addPrimaryKey("ID", "integer");
			}
			else {
				throw new SQLException(
						"Field ID is a reserved field which must be a primary field");
			}
		}
		String sqlStr = getCreateTableSql(tbl);
		con.executeSQL(sqlStr, false);
		con.release();
		return true;
	}

	/**
	 * Validates a table against the database
	 * 
	 * @param tbl
	 *            the table to be validated
	 * @return wheater the table validates or not
	 */
	public final Boolean validateTable(Table tbl) throws SQLException,
			ConnectionException {
		boolean isEqual = false;
		DatabaseConnection con = new DatabaseConnection(protocol, props,
				activeCon, unusedCon);
		Table dbtbl = getTableInfo(tbl.getName());
		con.release();
		if (dbtbl != null)
			isEqual = tbl.equals(dbtbl);
		// Returns true if the tables match
		return isEqual;
	}

	/**
	 * Dropps a table from the database
	 * 
	 * @return wheater the table was dropped successfully
	 */
	protected final Boolean dropDBTable(String sqlStr) throws SQLException,
			ConnectionException {
		DatabaseConnection con = new DatabaseConnection(protocol, props,
				activeCon, unusedCon);
		con.executeSQL(sqlStr, false);
		con.release();
		return true;
	}
}
