package coms.database;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

/**
 * An database connection which sql commands be be executed on.
 * 
 * @author Tom Arild Jakobsen & Mats Mortensen
 */
public class DatabaseConnection {
	private long tzero;
	private Connection connection;
	// Connection pools
	private Set<DatabaseConnection> activeCon;
	private Set<DatabaseConnection> unusedCon;

	/**
	 * Creates a database connection
	 * 
	 * @throws ConnectionException
	 *             tells why the connection to the database failed
	 */
	public DatabaseConnection(String protocol, Properties props,
			Set<DatabaseConnection> active, Set<DatabaseConnection> unused)
			throws ConnectionException {
		try {
			// Get a connection to the database
			connection = DriverManager.getConnection(protocol, props);
			// Connection pools
			activeCon = active;
			unusedCon = unused;
		}
		catch (Exception e) {
			e.printStackTrace();
			throw new ConnectionException(e.getMessage());
		}
		// Sets the timestamp for this connection
		tzero = System.currentTimeMillis();
	}

	/**
	 * Releases the connection to the connection pool
	 */
	public void release() {
		activeCon.remove(this);
		unusedCon.add(this);
	}

	/**
	 * Closes the Connection to the database
	 */
	public boolean close() {
		try {
			connection.close();
			return true;
		}
		catch (SQLException e) {
			// If it doesn't close, no biggy
			return false;
		}
	}

	/**
	 * Gets metadata information concerning database
	 * 
	 * @return a DatabaseMetaData object
	 * @throws an
	 *             SQLException
	 */
	public DatabaseMetaData getMetaData() throws SQLException {
		return connection.getMetaData();
	}

	/**
	 * Executes a SQL string on the connection
	 * 
	 * @param sqlStr
	 *            the SQL string to be executed
	 * @return a Set of Maps where each Map contains a table row
	 * @throws an
	 *             PersistentException if there is a problem executing the SQL
	 *             string
	 */
	public List<Map<String, Object>> executeSQL(String sqlStr,
			boolean returnValue) throws SQLException {
		return executeSQL(sqlStr, new Object[0], returnValue);
	}

	/**
	 * Executes a SQL string on the connection
	 * 
	 * @param sqlStr
	 *            the SQL string to be executed
	 * @param params
	 *            parameters to be put in the SQL string
	 * @return a Set of Maps where each Map contains a table row
	 * @throws a
	 *             PersistentException if there is a problem executing the SQL
	 *             string
	 */
	public List<Map<String, Object>> executeSQL(String sqlStr, Object[] params,
			boolean returnValue) throws SQLException {
//		System.out.println("executeSQL... " + sqlStr);
		String label;
		Map<String, Object> tmp;
		List<Map<String, Object>> result = new ArrayList<Map<String, Object>>();
		// Prepares the SQL statement
		PreparedStatement pstmt = connection.prepareStatement(sqlStr,
				ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_UPDATABLE);
		// Sets the parameters into the string from the params array
		for (int j = 0; j < params.length; j++) {
			pstmt.setObject(j + 1, params[j]);
		}
		if (returnValue) {
			// Executes the query and gets the metadata
			ResultSet rs = pstmt.executeQuery();
			ResultSetMetaData rsmd = rs.getMetaData();
			// Puts the metadata into an easier handable state to return
			while (rs.next()) {
				tmp = new HashMap<String, Object>();
				for (int i = 1; i < rsmd.getColumnCount() + 1; i++) {
					label = rsmd.getColumnLabel(i).toLowerCase();
					// If no label (Max/Min functions used?) set default label
					if (label.equals(""))
						label = "default";
					tmp.put(label, rs.getObject(i));
				}
				result.add(tmp);
			}
			rs.close();
		}
		else {
			pstmt.execute();
		}
		pstmt.close();
		connection.commit();
		return result;
	}

	/**
	 * Calculates the amount of time the connection has been up
	 * 
	 * @return the time the connection has been up
	 */
	public Long getUptime() {
		return new Long(System.currentTimeMillis() - tzero);
	}
}
