/*
 * @(#)file      ComponentAgent.java
 * @(#)author    Mats Mortensen
 * @(#)version   0.9
 * 
 */
package coms.mbeans;

import java.io.IOException;
import java.net.MalformedURLException;
import javax.management.ObjectName;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;
import coms.CoMS;
import coms.core.Module;
import coms.metadata.MetaData;

/**
 * Class which primarily initializes the mbeanserver used to hold components It
 * also provides functions for registering and unregistering components
 */
public class ComponentAgent extends Module implements ComponentAgentMBean {
	private JMXConnectorServer cs;

	/**
	 * Registers a component to the mbeanserver with component domain.
	 * Associates a metadata object with the registered component. Initializes
	 * the database layer for the component Starts the component
	 * 
	 * @param data
	 *            A MetaData object containing meta information about the
	 *            component,
	 * @param componentClass
	 *            A Class object used to instantiate the component
	 */
	public void registerComponent(MetaData data, Class componentClass) {
		try {
			ObjectName componentName = createObjectName(CoMS
					.getComponentDomain(), data.getName(), data.getType());
			server.registerMBean(componentClass.newInstance(), componentName);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Unregisters a component from the mbeanserver
	 * 
	 * @param data
	 *            A MetaData object containing meta information about the
	 *            component
	 */
	public void unRegisterComponent(MetaData data) {
		try {
			ObjectName componentName = createObjectName(CoMS
					.getComponentDomain(), data.getName(), data.getType());
			server.unregisterMBean(componentName);
			log("Component unloaded: " + componentName);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Create a new RMI connector
	 * 
	 * @param listenURL
	 *            The URL the server should listen on
	 */
	public void createRMIConnector(String listenURL) {
		try {
			JMXServiceURL url = new JMXServiceURL(
					"service:jmx:rmi:///jndi/rmi://" + listenURL);
			cs = JMXConnectorServerFactory.newJMXConnectorServer(url, null,
					server);
		}
		catch (MalformedURLException e) {
			e.printStackTrace();
			System.exit(0);
		}
		catch (IOException e) {
			log("IOException: " + e.getMessage());
		}
	}

	/**
	 * Starts the RMI connector. Makes it listen for connections
	 * 
	 * @return Returns true if the RMI server was started correctly
	 */
	public Boolean startRMIServer() {
		try {
			cs.start();
			return true;
		}
		catch (IOException e) {
			log("Error: The RMI service could not be started. Is the rmiregistry running?");
		}
		return false;
	}

	/**
	 * Stops the RMI connector from listening for connections
	 */
	public void stopRMIServer() {
		try {
			cs.stop();
		}
		catch (IOException e) {
			e.printStackTrace();
			System.exit(0);
		}
	}
}
