/*
 * @(#)file      ComponentLoader.java
 * @(#)author    Mats Mortensen & Tom Arild Jakobsen
 * @(#)version   0.9
 * 
 */
package coms.mbeans;

import java.io.File;
import java.io.FilenameFilter;
import java.util.Arrays;
import java.util.Comparator;
import javax.management.ObjectName;
import coms.core.Module;

/**
 * This class is run as a thread and takes care of loading components from the
 * deployed folder.
 */
public class ComponentLoader extends Module implements ComponentLoaderMBean {
	// private Thread componentLoaderThread = null;
	// private String name = "ComponentLoader";
	private long lastModified;
	private long time;
	private String deployDir;

	/**
	 * Sets the directory the loader should look for components in
	 * 
	 * @param dir
	 *            The directory the loader should look for components in
	 */
	public void setDeployDir(String dir) {
		deployDir = dir;
	}

	/**
	 * Constructor Initialized last modified to a valid file modified value. If
	 * the value is too small the File.setLastModified() will fail.
	 */
	public ComponentLoader() {
		lastModified = new Long("1123770710000");
		time = 0;
	}

	/**
	 * Checks if a specific jar file in the deploy directory can be loaded A jar
	 * file will only be loaded if it was modified (i.e. was uploaded) after the
	 * last component that was loaded
	 * 
	 * @param file
	 *            A File object describing the jar file
	 */
	private void checkFile(File file) {
		long filemodified = file.lastModified();
		if (lastModified < filemodified) {
			deployJarFile(file);
			lastModified = filemodified;
		}
	}

	/**
	 * Uses Deployer MBean to deploy all components from a jar file
	 * 
	 * @param file
	 *            A File object describing the jar file
	 */
	private void deployJarFile(File file) {
		boolean success = loadJarFile(file);
		if (success)
			deployLoadedJars();
	}

	/**
	 * Calls the loadJarFile method in the Deployer. This method is called from
	 * the FileHandler.
	 * 
	 * @param file
	 *            A File object describing the jar file
	 * @return Returns true if the loading was successful
	 * @see coms.mbeans#Deployer Deployer
	 * @see coms.mbeans#FileHandler FileHandler
	 */
	public Boolean loadJarFile(File file) {
		try {
			ObjectName deployer = createObjectName("Deployer", "Core");
			return (Boolean) server.invoke(deployer, "loadJarFile",
					new Object[] { file },
					new String[] { File.class.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
		return null;
	}

	/**
	 * Will set a new timestamp to a file. Only called when a new jar package is
	 * added to the system.
	 * 
	 * @param file
	 *            A File object describing the jar file
	 * @see coms.mbeans#Deployer Deployer
	 */
	public Boolean _setJarTimeStamp(String name) {
		Boolean success = true;
		File jarfile = new File(deployDir, name);
		if (lastModified == 0)
			lastModified = new Long("1123770710000");
		time += 2000;
		lastModified += time;
		success = jarfile.setLastModified(lastModified);
		return success;
	}

	/**
	 * Calls the deployLoadedJars method in the Deployer to deploy all loaded
	 * jar files.
	 * 
	 * @return A string which specifying the result of the deployment
	 */
	public String deployLoadedJars() {
		try {
			time = 0;
			ObjectName deployer = createObjectName("Deployer", "Core");
			return (String) server.invoke(deployer, "deployLoadedJars", null,
					null);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
		return null;
	}

	public void loadComponents() {
		checkDeployFolder(new File(deployDir));
	}

	/**
	 * Checks the component directory for new packages. Loads packages in the
	 * same order as they were added.
	 * 
	 * @param file
	 *            A File object describing the directory
	 */
	private void checkDeployFolder(File dir) {
		// long filemodified = lastModified, tmp = 0;
		ExtensionFilter filter = new ExtensionFilter(".jar");
		// Get list of files in directory
		File[] children = dir.listFiles(filter);
		// Sort the file list based on lastmodified value
		// Oldest files first, leave out other file ext than .jar
		Arrays.sort(children, new FileModifiedComparator());
		// Check each file
		for (File fileEntry : children)
			checkFile(fileEntry);
	}
}

/**
 * A comparator which let a list of files be sorted based on their last modified
 * property
 */
class FileModifiedComparator implements Comparator<File> {
	public int compare(File file1, File file2) {
		long value1 = file1.lastModified();
		long value2 = file2.lastModified();
		if (value1 < value2)
			return -1;
		else if (value1 > value2)
			return 1;
		else
			return 0;
	}
}

/**
 * Class that specifies which files should be filtered when retrieving a file
 * list from a directory
 */
class ExtensionFilter implements FilenameFilter {
	private String extension;

	public ExtensionFilter(String extension) {
		this.extension = extension;
	}

	public boolean accept(File dir, String name) {
		return (name.endsWith(extension));
	}
}
