/*
 * @(#)file      ComponentManager.java
 * @(#)author    Mats Mortensen, Timo
 * @(#)version   0.9
 * 
 */
package coms.mbeans;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.management.Attribute;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import coms.CoMS;
import coms.core.Graph;
import coms.core.Module;
import coms.metadata.ComponentInfo;
import coms.metadata.MetaData;
import coms.metadata.MetaDataWriter;
import coms.metadata.Table;

/**
 * This class provides all methods necessary to remove and add components It is
 * used by the Deployer to register and run a component and provides several
 * methods that is used by the CoMS Admin GUI
 */
public class ComponentManager extends Module implements ComponentManagerMBean {
	// Mapping of components currently loaded
	private LinkedHashMap<String, MetaData> loadedList = new LinkedHashMap<String, MetaData>();
	// Mapping of packages currently deployed (with components belonging to each
	// package)
	private LinkedHashMap<String, ArrayList<String>> packageMap = new LinkedHashMap<String, ArrayList<String>>();
	// Mapping of packages that depends on a particular package.
	private LinkedHashMap<String, ArrayList<String>> packageDepMap = new LinkedHashMap<String, ArrayList<String>>();
	private String persistenceDir = "deploy/metapersistence/";

	public ComponentManager() {
		super();
	}

	/**
	 * Get a list of ObjectNames for currently loaded components
	 * 
	 * @return An iterator to traverse the list
	 */
//	TODO Remove me
	/*
	private Iterator getComponentObjectNamesIterator() {
		Set componentList = getComponentObjectNames();
		return componentList.iterator();
	}*/

	/**
	 * Get a list of ObjectNames for currently loaded components
	 * 
	 * @return An Set of objectnames
	 */
	public Set getComponentObjectNames() {
		ObjectName components = null;
		try {
			components = new ObjectName(CoMS.getComponentDomain() + "*:*");
			Set componentList = server.queryNames(components, null);
			return componentList;
		}
		catch (MalformedObjectNameException e) {
			e.printStackTrace(System.out);
			System.exit(3);
		}
		return null;
	}

	/**
	 * Register the component to the mbeanserver Invokes the method that
	 * actually does this in the ComponentAgent
	 * 
	 * @param data
	 *            MetaData object describing the component
	 * @param componentClass
	 *            The class object of the component
	 * @see coms.metadata#MetaData
	 */
	private void register(MetaData data, Class componentClass) {
		try {
			ObjectName componentAgent = createObjectName("ComponentAgent",
					"Core");
			server.invoke(componentAgent, "registerComponent", new Object[] {
					data, componentClass }, new String[] {
					MetaData.class.getName(), Class.class.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Unregister the component from the mbeanserver Invokes the method that
	 * actually does this in the ComponentAgent
	 * 
	 * @param data
	 *            MetaData object describing the component to be unregistered
	 */
	private void unRegister(MetaData data) {
		try {
			ObjectName componentAgent = createObjectName("ComponentAgent",
					"Core");
			server.invoke(componentAgent, "unRegisterComponent",
					new Object[] { data }, new String[] { MetaData.class
							.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Drops all the tables connected to the component
	 * 
	 * @param data
	 *            MetaData object describing the component
	 * @see coms.metadata#MetaData
	 */
	private void dropTables(MetaData data) {
		try {
			Iterator tlist = data.tableIterator();
			// if true we remove
			while (tlist.hasNext()) {
				Table table = (Table) tlist.next();
				// Calls the default database of the component to drop the
				// components tables
				ObjectName dbmbean = createObjectName(data.getDbType(), "Core");
				server.invoke(dbmbean, "dropTable", new Object[] { table },
						new String[] { Table.class.getName() });
			}
		}
		catch (Exception e) {
			String msg = "<span style=\"color:red\">Error: </span>Could not drop tables in database for component \""
					+ data.getName() + "\"";
			msg += "<br>Caused by: " + e.getMessage();
			log(msg);
		}
	}

	/**
	 * Will attemt to remove a jar package from the system
	 * 
	 * @param pkg
	 *            String specifying the unique package name
	 */
	public synchronized Boolean removePackage(String pkg) {
		try {
			Iterator iter;
			// If it was already removed return
			if (!packageMap.containsKey(pkg))
				return true;
			// Remove any packages that depends on this package
			if (packageDepMap.containsKey(pkg)) {
				// Get the packages that depends on the package to be removed
				ArrayList list = packageDepMap.get(pkg);
				iter = list.iterator();
				String pkey = "";
				// For each package that depends on the package to be removed
				while (iter.hasNext()) {
					pkey = (String) iter.next();
					removePackage(pkey);
				}
			}
			// Get the components for the package
			ArrayList clist = packageMap.get(pkg);
			iter = clist.iterator();
			// Unload the components from the system
			while (iter.hasNext()) {
				String comid = (String) iter.next();
				unloadComponent(comid);
			}
			// Remove this package from the package maps
			packageMap.remove(pkg);
			packageDepMap.remove(pkg);
			// Delete the package from the system
			deleteFromSystem(pkg);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
		return true;
	}

	/**
	 * Will unload a component from the system
	 * 
	 * @param comID
	 *            String specifying the unique component
	 */
	private void unloadComponent(String comID) {
		MetaData com, data;
		// If this component already is removed, return
		if (loadedList.containsKey(comID))
			data = loadedList.get(comID);
		else
			return;
		// Get the components that depends on the component to be removed
		List list = data.getDependOnMe();
		// Iterator it = list.iterator();
		while (true) {
			if (!list.isEmpty()) {
				com = (MetaData) list.get(0);
				list.remove(0);
			}
			else {
				break;
			}
			// If they are in the same package
			if (com.getFileName().equals(data.getFileName())) {
				data.removeDependsOnMe(com);
				unloadComponent(com.getID());
			}
		}
		removeComponent(data);
	}

	/**
	 * Removes a component from the system. Removal involves stopping the
	 * component thread, unregister it from MBeanServer, dropping its tables
	 * from the database and deleting its stored xml meta file.
	 * 
	 * @param data
	 *            MetaData describing the component
	 */
	public void removeComponent(MetaData data) {
		// If component exists
		if (data != null) {
			ObjectName com = createObjectName(CoMS.getComponentDomain(), data
					.getName(), data.getType());
			// Stops component thread
			try {
				server.invoke(com, "stop", null, null);
			}
			catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
			// Unregister from MBeanserver
			unRegister(data);
			// Drop its tables from the database
			dropTables(data);
			// Remove from loaded list
			loadedList.remove(data.getID());
			Graph dependencyGraph = null;
			try {
				// Get the dependency graph
				ObjectName deployer = createObjectName("Deployer", "Core");
				dependencyGraph = (Graph) server.getAttribute(deployer,
						"DependencyGraph");
			}
			catch (Exception e) {
				e.printStackTrace();
			}
			// Update the dependency graph
			dependencyGraph.removeNode(data);
			// Delete its stored xml config file if it exists
			// Make sure the paths have system dependent path separators
			String path = "metapersistence/" + data.getMetaFileName();
			path = path.replace('/', File.separatorChar);
			File savedConfig = new File(path);
			if (savedConfig.exists())
				deleteFromSystem(savedConfig.getAbsolutePath());
		}
	}

	/**
	 * Delete a jar file from the deploy folder. Calls the actual method in
	 * JarUploader
	 * 
	 * @param filename
	 *            String specifying the filename of the component
	 * @see coms.mbeans#FileHandler
	 */
	private void deleteFromSystem(String filename) {
		try {
			ObjectName comFileHandler = createObjectName("JarUploader", "Core");
			server.invoke(comFileHandler, "deleteJarFromSystem",
					new Object[] { filename }, new String[] { String.class
							.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Initializes a Component.
	 * 
	 * @param name
	 *            The ObjectName of the component to be initialized
	 * @param data
	 *            The metadata of the component
	 */
	private void initComponent(ObjectName name, MetaData data) {
		try {
			// Initialize web and ext dirs
			String serviceName = data.getService();
			String extDir = persistenceDir + serviceName + "/ext/";
			String webDir = "jetty/webapps/coms/services/" + serviceName;
			// Make sure the paths have system dependent path separators
			extDir = extDir.replace('/', File.separatorChar);
			webDir = webDir.replace('/', File.separatorChar);
			server.setAttribute(name, new Attribute("ExtDir", extDir));
			server.setAttribute(name, new Attribute("WebDir", webDir));
			// Initialize Component
			server.invoke(name, "initialize", new Object[] { data },
					new String[] { MetaData.class.getName() });
			// Initialize config options
			HashMap opt = data.getCustomOptions();
			if (opt != null) {
				if (!opt.isEmpty()) {
					Set set = opt.keySet();
					Iterator it = set.iterator();
					while (it.hasNext()) {
						String key = (String) it.next();
						String value = (String) opt.get(key);
						key = key.substring(0, 1).toUpperCase()
								+ key.substring(1);
						server.setAttribute(name, new Attribute(key, value));
					}
				}
			}
		}
		catch (ReflectionException e) {
			e.printStackTrace();
		}
		catch (Exception e) {
			System.out.println("name: " + name.toString());
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Starts the component
	 * 
	 * @param name
	 *            The ObjectName of the component
	 * @param data
	 *            The metadata of the component
	 */
	private void startComponent(ObjectName name, MetaData data) {
		try {
			// Start the component thread (only if its not a output/config
			// component)
			if (!data.getType().equalsIgnoreCase("output")
					&& !data.getType().equalsIgnoreCase("config")) {
				log("Starting " + data.getName() + " thread");
				server.invoke(name, "start", null, null);
			}
			else {
				log("Initiating " + data.getName());
			}
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Add a a component to the system
	 * 
	 * @param data
	 *            MetaData object describing the component,
	 * @see coms.metadata#MetaData
	 */
	public synchronized void addComponent(MetaData data) {
		ObjectName cname = createObjectName(CoMS.getComponentDomain(), data
				.getName(), data.getType());
		// Assign a ID to the component consisting of <service>.<com>
		data.setID(data.getService() + "." + data.getName());
		// Register the component with the mbeanserver
		register(data, data.getComponentClass());
		log("Component loaded: " + cname);
		System.out.println(cname + " loaded.");
		// Initialize component (sets Metadata for component)
		initComponent(cname, data);
		// Start the component thread (only if its not a output/config
		// component)
		startComponent(cname, data);
		// Mark it as activated
		data.setActivated(true);
		// Add it to loaded list
		loadedList.put(data.getID(), data);
		// Add it to the package map (maps packages to lists of components)
		if (packageMap.containsKey(data.getFileName())) {
			ArrayList<String> list = packageMap.get(data.getFileName());
			list.add(data.getID());
		}
		else {
			ArrayList<String> newlist = new ArrayList<String>();
			newlist.add(data.getID());
			packageMap.put(data.getFileName(), newlist);
		}
		// Get the dependency graph
		Graph dependencyGraph = null;
		try {
			ObjectName deployer = createObjectName("Deployer", "Core");
			dependencyGraph = (Graph) server.getAttribute(deployer,
					"DependencyGraph");
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		// Check if component is dependent on any of the packages currently
		// loaded
		Set keys = packageDepMap.keySet();
		Iterator it = keys.iterator();
		while (it.hasNext()) {
			String key = (String) it.next();
			// If key is not the component jar package
			if (!key.equals(data.getFileName())) {
				// If component is dependent on the package (key)
				if (dependencyGraph.isDependentOnPackage(data, key)) {
					ArrayList<String> list = packageDepMap.get(key);
					list.add(data.getFileName());
				}
			}
		}
		// If components package is not already in the map, add it
		if (!packageDepMap.containsKey(data.getFileName()))
			packageDepMap.put(data.getFileName(), new ArrayList<String>());
	}

	/**
	 * Returns a list of loaded components Used by the CoMS GUI
	 * 
	 * @return An ArrayList of objects describing components
	 */
	public synchronized LinkedHashSet<ComponentInfo> getComponentList() {
		// Build a set of ComponentInfo objects (contains less info than
		// Metadata objects)
		LinkedHashSet<ComponentInfo> comset = new LinkedHashSet<ComponentInfo>();
		Set keys = loadedList.keySet();
		Iterator it = keys.iterator();
		while (it.hasNext()) {
			String key = (String) it.next();
			MetaData data = (MetaData) loadedList.get(key);
			ComponentInfo info = new ComponentInfo(data.getID(),
					data.getName(), data.getVersion(), data.getType());
			info.setActivated(data.getActivated());
			comset.add(info);
		}
		return comset;
	}

	/**
	 * Returns a map of packages with components
	 * 
	 * @return A LinkedHashMap that maps each package to a list of components
	 */
	public synchronized LinkedHashMap<String, ArrayList<String>> getPackageList() {
		return packageDepMap;
	}

	/**
	 * Stops a component temporarily
	 * 
	 * @param comID
	 *            String identifying a component by its name
	 */
	public synchronized void stopComponent(String comID) {
		MetaData data = loadedList.get(comID);
		stopComponent(data);
	}

	/**
	 * Stops a component and all components that depends on it temporarily
	 * 
	 * @param data
	 *            MetaData object describing a component
	 */
	public synchronized void stopComponent(MetaData data) {
		ObjectName com = createObjectName(CoMS.getComponentDomain(), data
				.getName(), data.getType());
		// Get all components that depend on this component
		List list = Graph.allWhoDependOnMe(data);
		Iterator it = list.iterator();
		// Stop all components that depend on this component
		while (it.hasNext()) {
			MetaData component = (MetaData) it.next();
			ObjectName comName = createObjectName(CoMS.getComponentDomain(),
					component.getName(), component.getType());
			if (component.getActivated()) {
				try {
					server.invoke(comName, "stop", null, null);
					component.setActivated(false);
					log("Component: \"" + component.getName()
							+ "\" has been suspended");
				}
				catch (Exception e) {
					e.printStackTrace();
					System.exit(0);
				}
			}
		}
		// Stop this component
		if (data.getActivated()) {
			try {
				server.invoke(com, "stop", null, null);
			}
			catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
			data.setActivated(false);
			log("Component: \"" + data.getName() + "\" has been suspended");
		}
	}

	/**
	 * Start a component that has previously been suspended
	 * 
	 * @param comID
	 *            String identifying a unique component
	 */
	public synchronized void startComponent(String comID) {
		MetaData data = loadedList.get(comID);
		startComponent(data);
	}

	/**
	 * Starts a component and all components that it depends on that has
	 * previously been suspended.
	 * 
	 * @param data
	 *            MetaData object describing a component
	 */
	public synchronized void startComponent(MetaData data) {
		ObjectName com = createObjectName(CoMS.getComponentDomain(), data
				.getName(), data.getType());
		// Get all components this component depend on
		List list = Graph.allWhoIDependOn(data);
		Iterator it = list.iterator();
		// Starts all components that this component depend on
		while (it.hasNext()) {
			MetaData component = (MetaData) it.next();
			ObjectName comName = createObjectName(CoMS.getComponentDomain(),
					component.getName(), component.getType());
			if (!component.getActivated()) {
				try {
					server.invoke(comName, "start", null, null);
					component.setActivated(true);
					log("Component: \"" + component.getName()
							+ "\" has been resumed");
				}
				catch (Exception e) {
					e.printStackTrace();
					System.exit(0);
				}
			}
		}
		// Start this component
		if (!data.getActivated()) {
			try {
				server.invoke(com, "start", null, null);
				data.setActivated(true);
				log("Component: \"" + data.getName() + "\" has been resumed");
			}
			catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
		}
	}

	/**
	 * Gets component configuration options
	 * 
	 * @param comID
	 *            String identifying a unique component
	 * @return A LinkedHashMap with a component's configuration options
	 */
	public synchronized HashMap<String, String> _getComponentConfig(String comID) {
		ObjectName name = null;
		try {
			MetaData data = loadedList.get(comID);
			// Update config options from component variables
			HashMap<String, String> opt = data.getCustomOptions();
			HashMap<String, String> newopt = new HashMap<String, String>();
			name = createObjectName(CoMS.getComponentDomain(), data.getName(),
					data.getType());
			if (opt != null) {
				if (!opt.isEmpty()) {
					Set set = opt.keySet();
					Iterator it = set.iterator();
					while (it.hasNext()) {
						String key = (String) it.next();
						key = key.substring(0, 1).toUpperCase()
								+ key.substring(1);
						String value = (String) server.getAttribute(name, key);
						newopt.put(key, value);
					}
					opt = newopt;
				}
			}
			return opt;
		}
		catch (Exception e) {
			System.out.println("name: " + name.toString());
			e.printStackTrace();
			System.exit(0);
		}
		return null;
	}

	/**
	 * Gets component status (activated/deactivated)
	 * 
	 * @param comID
	 *            String identifying a unique component
	 * @return True if the component is activated
	 */
	public synchronized Boolean _getComponentStatus(String comID) {
		MetaData data = loadedList.get(comID);
		return (Boolean) data.getActivated();
	}

	/**
	 * Sets component status (activated/deactivated)
	 * 
	 * @param comID
	 *            String identifying a unique component
	 * @param status
	 *            Set the activation status of the component to true or false
	 */
	public synchronized void _setComponentStatus(String comID, Boolean status) {
		MetaData data = loadedList.get(comID);
		data.setActivated(status);
	}

	/**
	 * Changes the configuration for a component and saves it to the components
	 * metafile
	 * 
	 * @param comID
	 *            String identifying a unique component
	 * @param param
	 *            A updated LinkedHashMap with all the components config options
	 *            which replaces the old one.
	 * @return Returns true if the change was successful
	 */
	public synchronized Boolean changeComponentConfig(String comID,
			HashMap<String, String> param) {
		ObjectName name = null;
		try {
			MetaData data = loadedList.get(comID);
			data.setCustomOptions(param);
			name = createObjectName(CoMS.getComponentDomain(), data.getName(),
					data.getType());
			// Update component variables
			Map opt = data.getCustomOptions();
			if (opt != null) {
				if (!opt.isEmpty()) {
					Set set = opt.keySet();
					Iterator it = set.iterator();
					while (it.hasNext()) {
						String key = (String) it.next();
						String value = (String) opt.get(key);
						key = key.substring(0, 1).toUpperCase()
								+ key.substring(1);
						server.setAttribute(name, new Attribute(key, value));
					}
				}
			}
			String serviceName = data.getService();
			File dir = new File("deploy/metapersistence/" + serviceName + "/");
			MetaDataWriter.write(data, dir);
			return true;
		}
		catch (Exception e) {
			System.out.println("name: " + name.toString());
			e.printStackTrace();
			System.exit(0);
		}
		return false;
	}
}
