/*
 * @(#)file      Deployer.java
 * @(#)author    Mats Mortensen
 * @(#)version   0.9
 * 
 */
package coms.mbeans;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Set;
import javax.management.MBeanServer;
import javax.management.MBeanServerConnection;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import coms.CoMS;
import coms.core.DeploymentException;
import coms.core.Graph;
import coms.core.JarLoader;
import coms.core.JarLoaderException;
import coms.core.JarValidityException;
import coms.core.Module;
import coms.metadata.Bind;
import coms.metadata.DeploymentDescWriter;
import coms.metadata.DeploymentDescription;
import coms.metadata.DeploymentDescriptorParser;
import coms.metadata.MetaData;
import coms.metadata.MetaDataParser;
import coms.metadata.MetaDataWriter;
import coms.metadata.ServiceDescription;
import coms.metadata.Table;

/**
 * This class provides all methods necessary to be ready to deploy all
 * components in a jar package. It will do all preliminary checks before it uses
 * the ComponentManager to actually load each component into the system. It
 * comes in to play after the component loader has loaded a component jar file
 * or when a new jar file is added through the Admin GUI
 */
public class Deployer extends Module implements DeployerMBean {
	private LinkedHashMap<String, MetaData> metadataMap = new LinkedHashMap<String, MetaData>();
	private LinkedHashMap<String, MetaData> loadedList;
	private LinkedHashMap<String, MetaData> deployList = new LinkedHashMap<String, MetaData>();
	private ArrayList<ServiceDescription> loadedServiceDescList = new ArrayList<ServiceDescription>();
	private ArrayList<String> loadedPackages = new ArrayList<String>();
	private Graph dependencyGraph = new Graph();
	private String deployDir = "deploy/";
	private String persistenceDir = deployDir + "metapersistence/";
	private String componentWebDir = "jetty/webapps/coms/services/";
	JarLoader jarLoader;
	//TODO Remove me
	//private final Class[] parameters = new Class[] { URL.class };

	/**
	 * Makes the dependency graph available to other modules
	 * 
	 * @return A dependency graph object
	 */
	public Graph getDependencyGraph() {
		return dependencyGraph;
	}

	/* Defines a class and recursively resolves class dependencies */
	private void defineClass(ClassLoader cl, Method method, String key, HashMap<String, byte[]> classes) throws JarLoaderException{
		for(int i=0;i<1;i++){
			try{	
				// Check if already defined
				try{
					Class.forName(key);
					return;
				}catch(Exception e){}
				
				byte[] bytes = (byte[]) classes.get(key);
				method.invoke(cl,new Object[]{null, bytes, 0,  bytes.length});
			}catch(Exception e){
				String depclass = "";
				try{
					String[] parts = e.getCause().toString().split(" ");
					depclass = parts[1].replace('/', '.');
				}catch(Exception e2){
					// Ignore this one
				}
				
				// Class had dependencies-->try loading class it depended on
				if(classes.containsKey(depclass)){
					defineClass(cl, method, depclass, classes);
					// Make sure we run method again to load original class
					i--;
				}else{
					throw new JarLoaderException("Could not find class: "+depclass);
				}
				
			}	
		}
	}
	
	/* Defines all classes in a hashmap */
	private void defineClasses(ClassLoader cl, Method method, HashMap<String, byte[]> classes) throws JarLoaderException{
	
		if(classes.size() == 0)
			return;
			
		Iterator it = classes.keySet().iterator();
		String key = null;
		byte[] bytes = null;
		while(it.hasNext()){
			key = (String) it.next();
			defineClass(cl, method, key, classes);
		}
	}	

	private void loadClassesFromJar(File file)throws JarLoaderException {
		try {
			// Create jar loader
			jarLoader = new JarLoader(file);
			// Open the jar file
			jarLoader.open();
			// Get system class loader
			ClassLoader sysloader = ClassLoader.getSystemClassLoader();
			Class sysclass = ClassLoader.class;
			// Use a "trick" to access its protected defineClass method
			Method method = sysclass.getDeclaredMethod("defineClass",
					new Class[] { String.class, byte[].class, int.class,
							int.class });
			method.setAccessible(true);
			// Read all classes from jar file
			jarLoader.readAll();
			// Get all classes found
			HashMap<String, byte[]> classes = jarLoader.getClasses();
			// Close jar
			jarLoader.close();
			// Use system class loader to define all classes from jar file
			// Obs! dependencies problem
			defineClasses(sysloader, method, classes);

		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Loads all information about a jar package into the system. Only when all
	 * this information is available all dependency checks can be performed.
	 * 
	 * @param file
	 *            A File object describing the jar package to be loaded
	 * @return Returns true if the loading was successful
	 */
	public Boolean loadJarFile(File file) {
		try {
			// New loaded list
			loadedList = new LinkedHashMap<String, MetaData>();
			// Load classes from jar file
			loadClassesFromJar(file);
			// Create jar loader
			jarLoader = new JarLoader(file);
			// Open the jar file
			jarLoader.open();
			DeploymentDescription deployDesc = null;
			// Check if the file "DeploymentDescriptor.xml" exists
			if (jarLoader.fileExists("DeploymentDescriptor.xml")) {
				// The deployment descriptor is parsed and returnes a deployment
				// description object
				DeploymentDescriptorParser parser = new DeploymentDescriptorParser();
				deployDesc = parser
						.parse(jarLoader, "DeploymentDescriptor.xml");
				loadedList.putAll(deployDesc.getComponentMap());
			}
			else {
				String msg = "The file \"DeploymentDescriptor.xml\" was not";
				msg += " found in the jar package \"" + file.getName() + "\"";
				throw new DeploymentException(msg);
			}
			// Check if meta persistence files should be used
			boolean useMetaPersistence = canUseMetaPersistence(deployDesc);
			// Get the metadata for each component
			getMetaDataForComponents(file, deployDesc, useMetaPersistence);
			// Check if all the necessary get/set methods are present in the
			// component
			checkConfigMethods();
			// Add the service desc to list of loaded service descriptions
			loadedServiceDescList.add(deployDesc.getServiceDescription());
			// Copy the ext and web folders to their destinations
			copyComponentFolders(deployDesc);
			// Close jar file
			jarLoader.close();
			try {
				// Use CoMS class loader to keep repository of component classes
				String serviceName = deployDesc.getServiceDescription()
						.getService().getName();
				ObjectName ccl = createObjectName("ComponentClassRepository",
						"Core");
				server.invoke(ccl, "loadFromJar", new Object[] { file,
						serviceName }, new String[] { File.class.getName(),
						String.class.getName() });
			}
			catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
			// Copy the loaded component mappings to the deploy list
			deployList.putAll(loadedList);
			// Add to loaded packages
			loadedPackages.add(file.getName());
			return true;
		}
		catch (JarLoaderException e) {
			String msg = "<span style=\"color:red\">Error: </span>Could not load component from \""
					+ file.getName() + "\"";
			msg += "<br>Caused by: " + e.getMessage();
			log(msg);
			if (jarLoader != null)
				try {
					jarLoader.close();
				}
				catch (JarLoaderException e2) {
					e2.printStackTrace();
					System.exit(0);
				}
			deleteFromSystem(file);
			return false;
		}
		catch (DeploymentException e) {
			String msg = "<span style=\"color:red\">Error: </span>Could not load components from \""
					+ file.getName() + "\"";
			msg += "<br>Caused by: " + e.getMessage();
			log(msg);
			try {
				jarLoader.close();
			}
			catch (JarLoaderException e2) {
				e2.printStackTrace();
				System.exit(0);
			}
			deleteFromSystem(file);
			return false;
		}
		catch (JarValidityException e) {
			String msg = "<span style=\"color:red\">Error: </span>Could not load components from \""
					+ file.getName() + "\"";
			msg += "<br>Caused by: " + e.getMessage();
			log(msg);
			try {
				jarLoader.close();
			}
			catch (JarLoaderException e2) {
				e2.printStackTrace();
				System.exit(0);
			}
			deleteFromSystem(file);
			return false;
		}
		catch (SAXParseException e) {
			// well-formedness error
			String msg = "<span style=\"color:red\">Error: </span>Could not load components from \""
					+ file.getName() + "\"";
			msg += "<br>Caused by: " + e.getMessage() + " at line "
					+ e.getLineNumber() + ", column " + e.getColumnNumber();
			log(msg);
			try {
				jarLoader.close();
			}
			catch (JarLoaderException e2) {
				e2.printStackTrace();
				System.exit(0);
			}
			deleteFromSystem(file);
			return false;
		}
		catch (SAXException e) {
			String msg = "<span style=\"color:red\">Error: </span>Could not load components from \""
					+ file.getName() + "\"";
			msg += "<br>Caused by: " + e.getMessage();
			log(msg);
			try {
				jarLoader.close();
			}
			catch (JarLoaderException e2) {
				e2.printStackTrace();
				System.exit(0);
			}
			deleteFromSystem(file);
			return false;
		}
	}

	public void copyComponentFolders(DeploymentDescription deployDesc)
			throws JarLoaderException {
		// Copy the ext and web folders to their destinations
		String serviceName = deployDesc.getServiceDescription().getService()
				.getName();
		
		// If component web dir does not exist, create it
		File cwebdir = new File(componentWebDir);
		if(!cwebdir.exists())
			cwebdir.mkdir();	
				
		String wdir = componentWebDir + serviceName + "/";
		wdir = wdir.replace('/', File.separatorChar);
		File webDir = new File(wdir);
		String edir = persistenceDir + serviceName + "/";
		edir = edir.replace('/', File.separatorChar);
		File extDir = new File(edir);
		if (!webDir.exists())
			webDir.mkdir();
		if (!extDir.exists())
			extDir.mkdir();
		String exdir = edir + "/ext/";
		exdir = exdir.replace('/', File.separatorChar);
		extDir = new File(exdir);
		if (!extDir.exists())
			extDir.mkdir();
		jarLoader.copyFolder("web", webDir.getPath());
		jarLoader.copyFolder("ext", extDir.getPath());
	}

	private void setupBindings() throws Exception {
		Iterator iter = deployList.keySet().iterator();
		while (iter.hasNext()) {
			String key = (String) iter.next();
			MetaData data = (MetaData) deployList.get(key);
			// Initialize bindings
			Set bindings = data.getBindings();
			Iterator it = bindings.iterator();
			String host, cls;
			Class[] constArgs = new Class[] { MBeanServer.class, String.class };
			while (it.hasNext()) {
				Bind binding = (Bind) it.next();
				MBeanServer mbs = null;
				// Get correct mbean server
				if (binding.getComp().indexOf(":") == -1) {
					// Using local server
					mbs = server;
				}
				else {
					String[] parts = binding.getComp().split(":");
					host = parts[0];
					cls = parts[1];
					// Using external server
					JMXConnector jmxc = JMXConnectorFactory
							.connect(new JMXServiceURL(
									"service:jmx:rmi:///jndi/rmi://" + host
											+ ":9999/server:" + cls));
					MBeanServerConnection mbsc = jmxc
							.getMBeanServerConnection();
					mbs = (MBeanServer) mbsc;
				}
				// Create listener
				Class cl = Class.forName(binding.getBindingClass());
				Constructor cons = cl.getConstructor(constArgs);
				// Get "to" component
				MetaData md = deployList.get(binding.getTo());
				ObjectName componentName = createObjectName(CoMS
						.getComponentDomain(), md.getName(), md.getType());
				ObjectName bindingcomp = createObjectName(CoMS
						.getComponentDomain(), data.getName(), data.getType());
				Object listener = cons.newInstance(new Object[] { mbs,
						bindingcomp.toString() });
				// Add listener
				server.invoke(componentName, "addNotificationListener",
						new Object[] { listener, listener, null },
						new String[] { NotificationListener.class.getName(),
								NotificationFilter.class.getName(),
								Object.class.getName() });
			}
		}
	}

	/**
	 * Deploys all components described in the deploylist
	 * 
	 * @return Returns a string with result messages
	 */
	public String deployLoadedJars() {
		try {
			// Add the component metadata objects to the metadata mapping
			metadataMap.putAll(deployList);
			// Build dependency graph
			buildDependencyGraph();
			// Check all components in package for uniqueness against system
			checkUniqueInSystem();
			// Check for circular dependencies
			checkCircularDependencies();
			// Deploy components
			deployComponents();
			// Deploy all services
			deployServices();
			// Setup all component bindings
			setupBindings();
			return "All packages were deployed successfully";
		}
		catch (JarValidityException e) {
			String msg = "<span style=\"color:red\">Error: </span>Could not load components from jar package";
			msg += "<br>Caused by: " + e.getMessage();
			rollbackDependencyGraph();
			log(msg);
			deletePackages();
		}
		catch (DeploymentException e) {
			String msg = "<span style=\"color:red\">Error: </span>Could not load components from jar package";
			msg += "<br>Caused by: " + e.getMessage();
			log(msg);
			try {
				rollbackDeploy();
			}
			catch (Exception e2) {
				// Unrecoverable
				e2.printStackTrace();
				System.exit(0);
			}
			deletePackages();
		}
		catch (Exception e) {
			// Should not happen
			e.printStackTrace();
			System.exit(0);
		}
		finally {
			// Create new deployment list
			deployList = new LinkedHashMap<String, MetaData>();
			// Create new package list
			loadedPackages = new ArrayList<String>();
			// Create new ServiceDescription list
			loadedServiceDescList = new ArrayList<ServiceDescription>();
		}
		return "Error";
	}

	/**
	 * Restores the dependency graph to the point before packages were tried to
	 * be deployed
	 */
	private void rollbackDependencyGraph() {
		Iterator it = deployList.values().iterator();
		while (it.hasNext()) {
			MetaData com = (MetaData) it.next();
			dependencyGraph.removeNode(com);
			metadataMap.remove(com);
		}
	}

	/**
	 * Deletes all loaded packages
	 */
	private void deletePackages() {
		Iterator it = loadedPackages.iterator();
		while (it.hasNext())
			deleteFromSystem(new File(deployDir, (String) it.next()));
	}

	/**
	 * Deploys all loaded services
	 */
	private void deployServices() throws DeploymentException {
		Iterator it = loadedServiceDescList.iterator();
		while (it.hasNext()) {
			ServiceDescription servDesc = (ServiceDescription) it.next();
			String error = addServiceDescription(servDesc);
			if (!error.equals(""))
				throw new DeploymentException(error);
		}
	}

	/**
	 * Deploys all loaded components
	 */
	private void deployComponents() throws DeploymentException, Exception {
		Iterator it = deployList.values().iterator();
		while (it.hasNext()) {
			MetaData data = (MetaData) it.next();
			deploy(data);
		}
	}

	/**
	 * Checks if any components have circular dependencies
	 */
	private void checkCircularDependencies() throws JarValidityException {
		Iterator it = metadataMap.values().iterator();
		while (it.hasNext())
			if (dependencyGraph.hasCycle((MetaData) it.next()))
				throw new JarValidityException(
						"The components have circular dependencies");
	}

	/**
	 * Builds a dependency graph of metadata objects
	 */
	private void buildDependencyGraph() throws JarValidityException {
		Iterator it = deployList.values().iterator();
		while (it.hasNext()) {
			MetaData com = (MetaData) it.next();
			Iterator dlist = com.getDependencies();
			;
			while (dlist.hasNext()) {
				String depOn = (String) dlist.next();
				MetaData data = (MetaData) metadataMap.get(depOn);
				if (data == null) {
					String msg = "Component: " + com.getName()
							+ " has a component listed as ";
					msg += " a dependency which does not exist in the package or the system";
					throw new JarValidityException(msg);
				}
				com.addDependency(data);
			}
			dependencyGraph.addNode(com);
		}
	}

	/**
	 * Checks if component names and table names from the components in the
	 * deploy list are unique compared to what is already loaded in the system.
	 */
	private void checkUniqueInSystem() throws JarValidityException, Exception {
		Iterator it = deployList.values().iterator();
		while (it.hasNext()) {
			MetaData com = (MetaData) it.next();
			// Check if name is unique
			if (!isUnique(com)) {
				String msg = "The component name \"" + com.getName()
						+ "\" already exist";
				throw new JarValidityException(msg);
			}
			// Check if table names are unique
			if (!isSQLTablesUnique(com)) {
				String msg = "The component \"" + com.getName()
						+ "\" has table names which already exist";
				throw new JarValidityException(msg);
			}
		}
	}

	/**
	 * Checks if the methods for configuring the component is present in the
	 * component
	 */
	private void checkConfigMethods() throws JarValidityException,
			DeploymentException {
		HashMap<String, String> customOpt;
		Method[] methods;
		Class componentClass;
		// For each component to be deployed
		Iterator it = loadedList.values().iterator();
		while (it.hasNext()) {
			MetaData com = (MetaData) it.next();
			try {
				// Get component class
				componentClass = Class.forName(com.getClassName());
				methods = componentClass.getDeclaredMethods();
				customOpt = com.getCustomOptions();
				// If component has custom options
				if (customOpt != null) {
					// For each custom option
					Iterator it2 = customOpt.keySet().iterator();
					while (it2.hasNext()) {
						String metname = (String) it2.next();
						metname = metname.substring(0, 1).toUpperCase()
								+ metname.substring(1);
						String getname = "get" + metname;
						String setname = "set" + metname;
						boolean foundget = false;
						boolean foundset = false;
						// Check if component has get/set methods corresponding
						// to the option
						for (int i = 0; i < methods.length; i++) {
							if (methods[i].getName().compareTo(getname) == 0)
								foundget = true;
							if (methods[i].getName().compareTo(setname) == 0)
								foundset = true;
						}
						// If both arn't found
						if (!foundget) {
							String msg = "<span style=\"color:red\">Error: </span>Could not load component \""
									+ com.getName() + "\"";
							msg += "<br>Caused by: Couldn't find the method \""
									+ getname + "\" in \"" + com.getClassName()
									+ "\"";
							throw new DeploymentException(msg);
						}
						else if (!foundset) {
							String msg = "<span style=\"color:red\">Error: </span>Could not load component \""
									+ com.getName() + "\"";
							msg += "<br>Caused by: Couldn't find the method \""
									+ setname + "\" in \"" + com.getClassName()
									+ "\"";
							throw new DeploymentException(msg);
						}
					}
				}
			}
			catch (ClassNotFoundException e) {
				String msg = "<span style=\"color:red\">Error: </span>Could not load component \""
						+ com.getName() + "\"";
				msg += "<br>Caused by: Class \"" + com.getClassName()
						+ "\" was not found in jar file";
				throw new DeploymentException(msg);
			}
		}
	}

	/**
	 * Checks if meta persistence files should be used based on version of
	 * deployment descriptor. If they cannot be used they will be deleted and
	 * the tables corresponding to the components will be deleted.
	 */
	private boolean canUseMetaPersistence(DeploymentDescription desc)
			throws JarValidityException, JarLoaderException, SAXException,
			SAXParseException, DeploymentException {
		try {
			// Create meta persistence dir if it does not exist
			String serviceName = desc.getServiceDescription().getService()
					.getName();
			File dir = new File(persistenceDir + serviceName + "/");
			if (!dir.exists())
				dir.mkdir();
			File deployDesc = new File(dir, "DeploymentDescriptor.xml");
			// Checks if there is a deployment descriptor file in the dir
			if (deployDesc.exists()) {
				// The deployment descriptor is parsed and returnes a deployment
				// description object
				DeploymentDescriptorParser parser = new DeploymentDescriptorParser();
				DeploymentDescription persistenceDesc = parser.parse(
						deployDesc, desc.getJarPackage());
				// If the already existing deployment description has a
				// different version than
				// the one found in the jar package, the persistence files
				// cannot be used
				String version = desc.getServiceDescription().getService()
						.getVersion();
				String pversion = persistenceDesc.getServiceDescription()
						.getService().getVersion();
				// Check if version is the same as the deployed jar
				if (version.equals(pversion)) {
					return true;
				}
				else {
					String msg = "XML meta files for a different version of the service "
							+ serviceName + " was ";
					msg += "already present in the system. ";
					msg += " This service's meta files and tables in the database will therefore be deleted";
					log(msg);
					// Find all the component tables and delete them from the
					// database
					LinkedHashMap components = persistenceDesc
							.getComponentMap();
					Set keyset = components.keySet();
					Iterator it = keyset.iterator();
					// Loop over all components
					while (it.hasNext()) {
						String key = (String) it.next();
						MetaData data = (MetaData) components.get(key);
						// Create new metadata parser
						MetaDataParser mdparser = new MetaDataParser(data, null);
						// If component persistent meta xml file exists
						File metafile = new File(dir, new File(data
								.getMetaFileName()).getName());
						if (metafile.exists()) {
							MetaData metadata = mdparser.parseFile(metafile);
							List tableList = metadata.getTables();
							Iterator iter = tableList.iterator();
							// Loop over all tables for component
							// and delete all existing tables
							while (iter.hasNext()) {
								Table table = (Table) iter.next();
								ObjectName componentAgent = createObjectName(
										data.getDbType(), "Core");
								server.invoke(componentAgent, "dropTable",
										new Object[] { table },
										new String[] { Table.class.getName() });
							}
						}
					}
					// Remove the directory and recreate it
					dir.delete();
					dir.mkdir();
					// Save the new deployment descriptor from the jar package
					// in the directory
					DeploymentDescWriter.write(desc, dir);
					return false;
				}
			}
			else {
				// Save the new deployment descriptor from the jar package in
				// the directory
				DeploymentDescWriter.write(desc, dir);
				return false;
			}
		}
		catch (Exception e) {
			String msg = "Could not drop tables used by old persistence meta files when loading: "
					+ desc.getJarPackage();
			msg += "<br>caused by: " + e.toString();
			throw new DeploymentException(msg);
		}
	}

	/**
	 * Reads, parses and gets the MetaData object for each component in the
	 * deploy list
	 */
	private void getMetaDataForComponents(File file,
			DeploymentDescription desc, boolean useMetaPersistence)
			throws JarValidityException, JarLoaderException, SAXException,
			SAXParseException, DeploymentException {
		MetaData data = null;
		String serviceName = desc.getServiceDescription().getService()
				.getName();
		File dir = new File(persistenceDir + serviceName + "/");
		try {
			Iterator it = loadedList.values().iterator();
			while (it.hasNext()) {
				// Metadata object to fill inn rest of meta information
				data = (MetaData) it.next();
				// Create new metadata parser
				MetaDataParser mdparser = new MetaDataParser(data, jarLoader);
				//
				if (useMetaPersistence) {
					// If there is saved information concerning this file
					File persistentMeta = new File(dir, data.getName() + ".xml");
					data = mdparser.parseFile(persistentMeta);
				}
				else {
					// Check if meta file for the component exists
					if (jarLoader.fileExists(data.getMetaFileName())) {
						// Parse meta file
						data = mdparser.parse(data.getMetaFileName());
						// Save a copy of the meta xml file in the persistence
						// directory
						MetaDataWriter.write(data, dir);
					}
					else {
						String msg = "The file \"" + data.getMetaFileName()
								+ "\" was not";
						msg += " found in the jar package \"" + file.getName()
								+ "\"";
						throw new DeploymentException(msg);
					}
				}
				// Saves component class in metadata
				Class componentClass = Class.forName(data.getClassName());
				data.setComponentClass(componentClass);
				// Save database type from deployment descriptor
				String type = desc.getServiceDescription().getService()
						.getDbType();
				data.setDbType(type);
				// Save name of service it belongs to
				data.setService(desc.getServiceDescription().getService()
						.getName());
			}
		}
		catch (ClassNotFoundException e) {
			String msg = "ClassNotFoundException: Could not load component \""
					+ data.getName() + "\"";
			msg += "<br>Caused by: Class \"" + data.getClassName()
					+ "\" was not found in jar file";
			throw new DeploymentException(msg);
		}
	}

	/**
	 * Uses the database handler to create a table in the databaes
	 * 
	 * @param table
	 *            Table object describing a table
	 * @param db
	 *            The name of the database which the table will be created in
	 */
	private boolean createTable(Table table, String db) {
		try {
			ObjectName mydb = createObjectName(db, "Core");
			return (Boolean) server.invoke(mydb, "createTable",
					new Object[] { table }, new String[] { Table.class
							.getName() });
		}
		catch (Exception e) {
			 e.printStackTrace();
			// System.exit(0);
			return false;
		}
	}

	/**
	 * Uses the database handler to check if a table exists in the database
	 * 
	 * @param table
	 *            Table object describing a table
	 * @param db
	 *            The name of the database which the table will be created in
	 */
	private boolean tableExists(Table table, String db) {
		try {
			ObjectName mydb = createObjectName(db, "Core");
			return (Boolean) server.invoke(mydb, "validateTable",
					new Object[] { table }, new String[] { Table.class
							.getName() });
		}
		catch (Exception e) {
			return false;
		}
	}

	/**
	 * Uses the database handler to drop a table from the database
	 * 
	 * @param table
	 *            Table object describing a table
	 * @param db
	 *            The name of the database which the table will be created in
	 */
	@SuppressWarnings("unused")
	private boolean dropTable(Table table, String db) {
		try {
			ObjectName mydb = createObjectName(db, "Core");
			return (Boolean) server.invoke(mydb, "dropTable",
					new Object[] { table }, new String[] { Table.class
							.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			// System.exit(0);
			return false;
		}
	}

	/**
	 * Creates tables for a component in the database, based on the components
	 * metadata
	 * 
	 * @param data
	 *            the metadata of a specific component
	 * @return wheater or not the tables could be created
	 */
	private void createTables(MetaData data) throws DeploymentException {
		Table table;
		try {
			Iterator tables = data.tableIterator();
			// For each table the component has...
			while (tables.hasNext()) {
				// ...create a table in the database
				table = (Table) tables.next();
				// Only create table if it does not already exist
				if (!tableExists(table, data.getDbType()))
					createTable(table, data.getDbType());
			}
		}
		catch (Exception e) {
			String msg = "<span style=\"color:red\">Error: </span>Could not create tables in database for component \""
					+ data.getName() + "\"";
			msg += "<br> Caused by: " + e.getMessage();
			msg += "<br>The component will therefore be removed from the system";
			throw new DeploymentException(msg);
		}
	}

	/**
	 * Deploys a component and the components it is dependent on recursively
	 * 
	 * @param com
	 *            MetaData object describing a component
	 * @see coms.metadata#MetaData
	 */
	private void deploy(MetaData com) throws DeploymentException, Exception {
		MetaData data;
		List dlist = com.getIDependOn();
		Iterator it = dlist.iterator();
		// If not already loaded
		if (!isLoaded(com)) {
			while (it.hasNext()) {
				data = (MetaData) it.next();
				deploy(data);
			}
			createTables(com);
			// Mark its jar package with timestamp
			if (setJarTimeStamp(com.getFileName())) {
				addToSystem(com);
			}
			else {
				String msg = "Could not change filemodified property on jar file: "
						+ com.getFileName();
				throw new DeploymentException(msg);
			}
		}
	}

	private Boolean setJarTimeStamp(String filename) {
		try {
			ObjectName cloader = createObjectName("ComponentLoader", "Core");
			return (Boolean) server.invoke(cloader, "_setJarTimeStamp",
					new Object[] { filename }, new String[] { String.class
							.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
		return null;
	}

	/**
	 * Rolls back the package deployment
	 */
	private void rollbackDeploy() throws Exception {
		// Unload all components
		Iterator it = deployList.values().iterator();
		while (it.hasNext()) {
			MetaData com = (MetaData) it.next();
			rollback(com);
		}
		// Rollback dependency graph
		rollbackDependencyGraph();
	}

	/**
	 * Recursively unloads a component and all the components it depends on
	 * 
	 * @param com
	 *            MetaData object describing a component
	 */
	private void rollback(MetaData com) throws Exception {
		Iterator dlist = com.getDependencies();
		int dependencyCount = iteratorCount(dlist);
		dlist = com.getDependencies();
		if (dependencyCount > 0) {
			while (dlist.hasNext()) {
				String depOn = (String) dlist.next();
				MetaData data = (MetaData) deployList.get(depOn);
				if (data != null)
					rollback(data);
			}
		}
		if (isLoaded(com))
			unloadComponent(com);
	}

	/**
	 * Checks if a component name is unique compared to the loaded components in
	 * the system
	 * 
	 * @param data
	 *            MetaData object describing the component
	 * @return Returns true if the component name is unique
	 */
	private boolean isUnique(MetaData data) throws Exception {
		boolean unique = true;
		Set list = getLoadedComponents();
		Iterator clist = list.iterator();
		// Check if the component name is unique compared to the other installed
		// components
		while (clist.hasNext()) {
			ObjectName component = (ObjectName) clist.next();
			MetaData componentData = (MetaData) server.getAttribute(component,
					"MetaData");
			// Check if name and version match
			if (data.getName().equals(componentData.getName()))
				unique = false;
		}
		return unique;
	}

	/**
	 * Check that the names for the sql tables isnt in use by a already loaded
	 * component
	 * 
	 * @param data
	 *            MetaData object describing a component
	 * @return True if no table names are in use, false othervise
	 */
	private boolean isSQLTablesUnique(MetaData data) throws Exception {
		Table table, ctable;
		boolean isUnique = true;
		String tablename = "", ctablename = "";
		MetaData componentData = null;
		Set list = getLoadedComponents();
		Iterator clist = list.iterator();
		Iterator tlist = data.tableIterator();
		// For each table the component has
		while (tlist.hasNext()) {
			table = (Table) tlist.next();
			tablename = table.getName();
			// For each installed component
			while (clist.hasNext()) {
				ObjectName component = (ObjectName) clist.next();
				componentData = (MetaData) server.getAttribute(component,
						"MetaData");
				Iterator it = componentData.tableIterator();
				// Check if unique against installed components tables
				while (it.hasNext()) {
					ctable = (Table) it.next();
					ctablename = ctable.getName();
					if (tablename.equals(ctablename)) {
						isUnique = false;
						break;
					}
				}
				if (isUnique == false)
					break;
			}
			if (isUnique == false)
				break;
		}
		if (!isUnique) {
			String msg = "<span style=\"color:red\">Error: </span> Could not load component \""
					+ data.getName() + "\"";
			msg += "<br>Caused by: The table name: " + tablename
					+ " already exists as part of an installed component.";
			log(msg);
		}
		return isUnique;
	}

	/**
	 * Adds the component specified to the CoMS system Invokes method
	 * "addComponent" in ComponentManager
	 * 
	 * @param data
	 *            A MetaData object containing meta information about a
	 *            component
	 * @see coms.mbeans#ComponentManager
	 */
	public void addToSystem(MetaData data) throws DeploymentException {
		try {
			ObjectName componentManager = createObjectName("ComponentManager",
					"Core");
			server.invoke(componentManager, "addComponent",
					new Object[] { data }, new String[] { MetaData.class
							.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Returns a list of loaded components
	 * 
	 * @return A lit of loaded components
	 */
	private Set getLoadedComponents() throws Exception {
		ObjectName comManager = createObjectName("ComponentManager", "Core");
		Set list = (Set) server
				.getAttribute(comManager, "ComponentObjectNames");
		return list;
	}

	/**
	 * Counts items in a list
	 * 
	 * @param it
	 *            An iterator to traverse the list
	 * @return An integer with the count
	 */
	private int iteratorCount(Iterator it) {
		int i = 0;
		while (it.hasNext()) {
			i++;
			it.next();
		}
		return i;
	}

	/**
	 * Returns true if a component is already loaded and false if it is not
	 * 
	 * @param comname
	 *            String identifying a component by its name
	 * @return Returns true if the component is loaded
	 */
	private boolean isLoaded(MetaData data) throws Exception {
		ObjectName com = createObjectName(CoMS.getComponentDomain(), data
				.getName(), data.getType());
		return server.isRegistered(com);
	}

	/**
	 * Deletes a jar file from the system
	 * 
	 * @param filename
	 *            String filename that specifies the file to be deleted
	 */
	private void deleteFromSystem(File file) {
		try {
			// Create jar loader
			jarLoader = new JarLoader(file);
			// Open the jar file
			jarLoader.open();
			DeploymentDescription deployDesc = null;
			// Check if the file "DeploymentDescriptor.xml" exists
			if (jarLoader.fileExists("DeploymentDescriptor.xml")) {
				// The deployment descriptor is parsed and returnes a deployment
				// description object
				DeploymentDescriptorParser parser = new DeploymentDescriptorParser();
				deployDesc = parser
						.parse(jarLoader, "DeploymentDescriptor.xml");
			}
			// Close jar file
			jarLoader.close();
			// Delete persistence directory
			String serviceName = deployDesc.getServiceDescription()
					.getService().getName();
			File dir = new File(persistenceDir + serviceName + "/");
			if (dir.exists()) {
				if (!deleteDir(dir))
					log("<span style=\"color:red\">Error: </span>Failed to delete (whole) directory: "
							+ dir.getPath());
			}
			// Delete jar file
			ObjectName comFileHandler = createObjectName("JarUploader", "Core");
			server.invoke(comFileHandler, "deleteJarFromSystem",
					new Object[] { file.getName() },
					new String[] { String.class.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	private boolean deleteDir(File dir) {
		if (dir.isDirectory()) {
			String[] children = dir.list();
			for (int i = 0; i < children.length; i++) {
				boolean success = deleteDir(new File(dir, children[i]));
				if (!success) {
					return false;
				}
			}
		}
		// The directory is now empty so delete it
		return dir.delete();
	}

	/**
	 * Uses the ServiceDescriptor module to deploy a new service description
	 * 
	 * @param serviceDesc
	 *            ServiceDescription object describing several services
	 * @see coms.mbeans#ServiceDescriptor
	 */
	private String addServiceDescription(ServiceDescription serviceDesc)
			throws DeploymentException {
		try {
			ObjectName descriptor = createObjectName("ServiceDescriptor",
					"Core");
			return (String) server.invoke(descriptor, "addServiceDescription",
					new Object[] { serviceDesc, metadataMap }, new String[] {
							ServiceDescription.class.getName(),
							HashMap.class.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
		return null;
	}

	/**
	 * Uses ComponentManager to remove a component from the system
	 * 
	 * @param com
	 *            MetaData object describing a component
	 * @see coms.mbeans#ComponentManager
	 */
	private void unloadComponent(MetaData com) {
		try {
			ObjectName comManager = createObjectName("ComponentManager", "Core");
			server.invoke(comManager, "removeComponent", new Object[] { com },
					new String[] { MetaData.class.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}
}
