package coms.mbeans;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import coms.core.Module;
import coms.core.Util;

/**
 * Lets components store and access files in the filesystem
 * 
 * @author Tom Arild Jakobsen & Mats Mortensen
 */
public class FileHandler extends Module implements FileHandlerMBean {
	private static final int DEFAULT_MAX_FILES = 100;
	private Integer maxFiles;

	/**
	 * Creates a FileStorage object
	 */
	public FileHandler() {
		this(DEFAULT_MAX_FILES);
	}

	/**
	 * Creates a FileStorage object
	 * 
	 * @param max
	 *            the maximum number of files in one folder
	 */
	public FileHandler(int max) {
		maxFiles = max;
		// If the filesystem folder don't exist, create it
		File file = new File("files");
		if (!file.exists())
			file.mkdir();
	}

	/**
	 * Sets the maximum number of allowed files in each folder
	 * 
	 * @param max
	 *            the new maximum number files in a folder
	 */
	public void setMaxFiles(Integer max) {
		maxFiles = max;
	}

	/**
	 * Gets the current maximum number files in each folder
	 * 
	 * @return the maximum number files in a folder
	 */
	public Integer getMaxFiles() {
		return maxFiles;
	}

	/**
	 * Takes a file and saves it on the same computer as the CoMS
	 * 
	 * @param stream
	 *            the input stream which the file will be transfered over
	 * @return the path to where the file is on disk
	 * @throws IOException
	 *             if there are problems with writing to the file
	 */
	public String putFile(InputStream stream) throws IOException {
		// Buffer of fixed size so that we don't take too much memory
		byte[] buf = new byte[500];
		int bytesReceived = 0;
		String path = getPath();
		// Make sure that file exists
		File file = new File(path);
		file.createNewFile();
		// Copy data via buffer
		FileOutputStream fos = new FileOutputStream(path);
		while ((bytesReceived = stream.read(buf)) > 0) {
			fos.write(buf, 0, bytesReceived);
		}
		stream.close();
		fos.close();
		return path;
	}

	/**
	 * Takes a file and saves it on the same computer as the CoMS
	 * 
	 * @param data
	 *            the data that is to be put in the file
	 * @return the path to where the file is on disk
	 * @throws IOException
	 *             if there are problems with writing to the file
	 */
	public String putFile(Byte[] data) throws IOException {
		String path = getPath();
		// Make sure that file exists
		File file = new File(path);
		file.createNewFile();
		// Copy data
		FileOutputStream fos = new FileOutputStream(path);
		fos.write(Util.byteObjToBytes(data));
		fos.close();
		return path;
	}

	/**
	 * Puts a Byte array into a file in the filesystem
	 * 
	 * @param fileId
	 *            the path of the file
	 * @param data
	 *            the data to be put into the file
	 * @throws IOException
	 *             if there is some problem in accessing/creating the file
	 */
	public void putIntoFile(String fileId, Byte[] data) throws IOException {
		File file = new File(fileId);
		file.createNewFile();
		// Copy data
		FileOutputStream fos = new FileOutputStream(fileId, true);
		fos.write(Util.byteObjToBytes(data));
		fos.close();
	}

	/**
	 * Accesses a file and sends its contents over an input stream
	 * 
	 * @param path
	 *            the path where the file is on disk
	 * @return an input stream from that file
	 * @throws FileNotFoundException
	 *             if the file is not found
	 */
	public InputStream readFile(String path) throws FileNotFoundException {
		File file = new File(path);
		// Returns the Files input stream
		if (file.exists()) {
			InputStream stream = new FileInputStream(path);
			return stream;
		}
		else {
			String msg = "Could not find file in path \"" + path + "\"";
			throw new FileNotFoundException(msg);
		}
	}

	/**
	 * Deletes a file from the disk
	 * 
	 * @param path
	 *            the path where the file to be deleted is on disk
	 * @throws FileNotFoundException
	 *             if the file is not found
	 */
	public void deleteFile(String path) {
		File file = new File(path);
		if (file.exists())
			file.delete();
	}

	/**
	 * Chooses a suitable place in the file space and makes an empty file
	 * 
	 * @return the path where the file can be saved
	 * @throws IOException
	 *             if the new file can't be created
	 */
	private String getPath() throws IOException {
		File filespace = new File("files");
		File[] folders = filespace.listFiles();
		String folderName = Integer.toString(folders.length);
		// Checks wheater there is room in one of the existing folders
		for (int i = 0; i < folders.length; i++) {
			if (folders[i].list().length < maxFiles) {
				folderName = folders[i].getName();
				break;
			}
		}
		// Creates handle to folder, if folder don't exist create one
		File folder = new File(filespace, folderName);
		if (!folder.exists())
			folder.mkdir();
		// Traverses the folder to find the first free name
		String[] files = folder.list();
		String fileName = Integer.toString(files.length);
		boolean found = false;
		for (int i = 0; i < maxFiles; i++) {
			for (int j = 0; j < files.length; j++) {
				if (files[j].equals(Integer.toString(i))) {
					found = true;
					break;
				}
			}
			if (!found) {
				fileName = Integer.toString(i);
				break;
			}
			found = false;
		}
		// Creates a new file at the first free space
		File file = new File(folder, fileName);
		if (file.exists()) {
			throw new RuntimeException("Return fileID that already exists!!");
		}
		file.createNewFile();
		return file.getAbsolutePath();
	}
}
