package coms.mbeans;

import java.io.IOException;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import coms.core.Config;
import coms.database.ConnectionException;
import coms.database.Database;
import coms.metadata.Table;
import coms.metadata.TableField;

/**
 * Gives out access to the Derby database in the form of DerbyConnections. Will
 * also monitor the Derby database
 * 
 * @author Tom Arild Jakobsen & Mats Mortensen
 */
public class HSQL extends Database implements HSQLMBean {
	private final String configUrl = "./config/database/HSQLConfig.xml";
	private final String driver = "org.hsqldb.jdbcDriver";
	private String protocol = "jdbc:hsqldb:HSQL/db_file";
	private Properties props;

	/**
	 * HSQL constructor
	 */
	public HSQL() {
		super();
		try {
			// Parses the configuration file
			// TODO endre hvor pathen begynner i ant fila?
			HashMap<String, String> config = Config.read(configUrl);
			// Create connection properties
			props = new Properties();
			protocol += config.get("name");
			// Initiates database
			init(driver, protocol, props);
		}
		catch (IOException e) {
			// Should not occur
			e.printStackTrace();
		}
		catch (Exception e) {
			// Should not occur
			e.printStackTrace();
		}
	}

	/**
	 * Uses the database to create a Table object to the corresponding name
	 * 
	 * @param name
	 *            the name of the table to get info about
	 * @return the complete metainfo of the table
	 */
	public Table getTableInfo(String table) throws SQLException,
			ConnectionException {
		// Selects the column names from a table in the database
		String sqlStr = "select c.columnname, c.columndatatype, c.columndefault from sys.syscolumns c, sys.systables t";
		sqlStr += " where  c.referenceid=t.tableid and t.tablename='" + table
				+ "'";
		return getTableDesc(table, sqlStr, "columnname", "columndatatype",
				"columndefault");
	}

	/**
	 * Adds a table to the database
	 * 
	 * @param tbl
	 *            The table to be added to the database
	 * @return wheater the table was added successfully
	 */
	protected String getCreateTableSql(Table tbl) throws SQLException,
			ConnectionException {
		String def;
		TableField primary = null;
		primary = tbl.getPrimaryKey();
		String sqlStr = "CREATE CACHED TABLE " + tbl.getName() + "(";
		// Add primary key
		sqlStr += primary.getName() + " " + toSqlType(primary.getType());
		sqlStr += " GENERATED BY DEFAULT AS IDENTITY (START WITH 1, INCREMENT BY 1) PRIMARY KEY, ";
		// Iterates over the fields in the table
		Iterator it = tbl.getFieldIterator();
		while (it.hasNext()) {
			TableField field = (TableField) it.next();
			// Adds all the non primary fields to the sql string
			if (!field.isPrimaryKey()) {
				sqlStr += field.getName() + " " + toSqlType(field.getType());
				// If there is a default value
				def = field.getDefault();
				if (def != null && !def.equals(""))
					sqlStr += " DEFAULT " + def;
				sqlStr += ", ";
			}
		}
		sqlStr += " PRIMARY KEY(" + primary.getName() + ")";
		sqlStr += ")";
		return sqlStr;
	}

	/**
	 * Dropps a table from the database
	 * 
	 * @return wheater the table was dropped successfully
	 */
	public Boolean dropTable(Table tbl) throws SQLException,
			ConnectionException {
		String sqlStr = "DROP TABLE " + tbl.getName();
		return dropDBTable(sqlStr);
	}

	/**
	 * Shut downs the database
	 */
	public void shutdown() throws SQLException, ConnectionException {
		String sqlStr = "SHUTDOWN";
		shutdownDB(sqlStr);
	}

	/**
	 * Converts a java type to a sql type
	 * 
	 * @param type
	 *            the java type to be converted
	 * @return the sql equivalent to the java type
	 */
	protected String toSqlType(String type) {
		if (type.equalsIgnoreCase("string"))
			return "VARCHAR(200)";
		else if (type.equalsIgnoreCase("bigdecimal"))
			return "DECIMAL";
		else if (type.equalsIgnoreCase("boolean"))
			return "CHAR FOR BIT DATA";
		else if (type.equalsIgnoreCase("byte"))
			return "CHAR";
		else if (type.equalsIgnoreCase("int"))
			return "INTEGER";
		else if (type.equalsIgnoreCase("integer"))
			return "INTEGER";
		else if (type.equalsIgnoreCase("short"))
			return "SMALLINT";
		else if (type.equalsIgnoreCase("double"))
			return "DOUBLE PRECISION";
		else if (type.equalsIgnoreCase("float"))
			return "DOUBLE PRECISION";
		else if (type.equalsIgnoreCase("date"))
			return "DATE";
		else if (type.equalsIgnoreCase("time"))
			return "TIME";
		else if (type.equalsIgnoreCase("timestamp"))
			return "TIMESTAMP";
		else
			return type;
	}
}
