package coms.mbeans;

import java.io.IOException;
import java.lang.management.ClassLoadingMXBean;
import java.lang.management.CompilationMXBean;
import java.lang.management.GarbageCollectorMXBean;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryMXBean;
import java.lang.management.MemoryManagerMXBean;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.OperatingSystemMXBean;
import java.lang.management.RuntimeMXBean;
import java.lang.management.ThreadMXBean;
import java.net.MalformedURLException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.LogManager;
import java.util.logging.LoggingMXBean;
import javax.management.Attribute;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.loading.MLet;
import javax.management.loading.MLetMBean;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;
import coms.config.ConfigOptions;
import coms.core.Module;
import coms.metadata.Table;

/**
 * This class is the starting point for loading and configuring system modules.
 * 
 * @author Mats Mortensen & Tom Arild Jakobsen
 */
public class ModuleAgent extends Module implements ModuleAgentMBean {
	private MLetMBean mletService;
	private String mletURL;
	private JMXConnectorServer cs;
	private ConfigOptions config;

	/**
	 * Sets the url location of the mlet file. Mlet file contains all modules to
	 * be loaded.
	 * 
	 * @param u
	 *            String
	 */
	public void setURL(String u) {
		mletURL = u;
	}

	/**
	 * Sets the config options for the system
	 * 
	 * @param c
	 *            Configuration options
	 */
	public void setConfig(ConfigOptions c) {
		config = c;
	}

	/**
	 * Configures module options
	 */
	private void configureModules() {
		try {
			// Get Inference Factory handle
			ObjectName factory = createObjectName("InferenceFactory", config
					.getModuleType());
			// Set the type of database the system uses
			server.setAttribute(factory, new Attribute("DBType", config
					.getDBType()));
			// Get JarUploader handle
			ObjectName jUploader = createObjectName("JarUploader", config
					.getModuleType());
			// Set the directory for saving components to
			server.setAttribute(jUploader, new Attribute("DeployDir", config
					.getDeployDir()));
			// Get component loader handle
			ObjectName comLoader = createObjectName("ComponentLoader", config
					.getModuleType());
			// Set the deployment directory for components
			server.setAttribute(comLoader, new Attribute("DeployDir", config
					.getDeployDir()));
			// Get component agent handle
			ObjectName comAgent = createObjectName("ComponentAgent", config
					.getModuleType());
//			TODO Remove me
			//ObjectName mlet = createObjectName("MLetLoader", config.getModuleType());
			// Start the component RMI server that is used for output components
			server.invoke(comAgent, "createRMIConnector", new Object[] { config
					.getComponentListenURL() }, new String[] { String.class
					.getName() });
			server.invoke(comAgent, "startRMIServer", null, null);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Loads the modules specified in a mlet file using the mlet service mbean
	 * 
	 * @param url
	 *            String of url
	 */
	private void loadModules(String url) throws Exception {
		Set loadedModules = mletService.getMBeansFromURL(url);
		Iterator iter = loadedModules.iterator();
		// Send to the log what modules where loaded
		while (iter.hasNext()) {
			Object o = iter.next();
			if (o instanceof ObjectInstance) {
				ObjectInstance oi = (ObjectInstance) o;
				log("Module loaded: " + oi.getObjectName().toString());
			}
		}
	}

	/**
	 * Registers a the mlet service mbean. The mlet service is used to load
	 * modules specified in a mlet file
	 */
	private void initMLet() throws Exception {
		mletService = new MLet();
		ObjectName name = createObjectName("MLetLoader", config.getModuleType());
		server.registerMBean(mletService, name);
	}

	/**
	 * Loads and configures all modules specified in the mlet file
	 */
	public void loadModules() throws Exception {
		// Load monitoring mbeans
		loadMonitorModules();
		// Start mlet service
		initMLet();
		// Load container modules
		loadModules(mletURL);
		// Configure container modules
		configureModules();
		// Create system database tables
		createSystemTables();
	}

	private void createSystemTables() throws Exception {
		ObjectName db = createObjectName(config.getDBType(), "Core");
		// Get system tables
		HashSet tables = config.getSystemTables();
		Iterator it = tables.iterator();
		// Create system tables
		while (it.hasNext()) {
			Table tbl = (Table) it.next();
			Boolean exists = (Boolean) server.invoke(db, "validateTable",
					new Object[] { tbl },
					new String[] { Table.class.getName() });
			if (!exists)
				server.invoke(db, "createTable", new Object[] { tbl },
						new String[] { Table.class.getName() });
		}
	}

	/**
	 * Loads the JVM monitoring mbeans
	 */
	private void loadMonitorModules() throws Exception {
		// Runtime Monitoring module
		ObjectName oname = new ObjectName(ManagementFactory.RUNTIME_MXBEAN_NAME);
		RuntimeMXBean rmxbean = ManagementFactory.getRuntimeMXBean();
		server.registerMBean(rmxbean, oname);
		// ClassLoading Monitoring module
		oname = new ObjectName(ManagementFactory.CLASS_LOADING_MXBEAN_NAME);
		ClassLoadingMXBean clmxbean = ManagementFactory.getClassLoadingMXBean();
		server.registerMBean(clmxbean, oname);
		// Compilation Monitoring module
		oname = new ObjectName(ManagementFactory.COMPILATION_MXBEAN_NAME);
		CompilationMXBean compmxbean = ManagementFactory.getCompilationMXBean();
		server.registerMBean(compmxbean, oname);
		// Memory Monitoring module
		oname = new ObjectName(ManagementFactory.MEMORY_MXBEAN_NAME);
		MemoryMXBean memmxbean = ManagementFactory.getMemoryMXBean();
		server.registerMBean(memmxbean, oname);
		// Thread Monitoring module
		oname = new ObjectName(ManagementFactory.THREAD_MXBEAN_NAME);
		ThreadMXBean threadmxbean = ManagementFactory.getThreadMXBean();
		server.registerMBean(threadmxbean, oname);
		// OS Monitoring module
		oname = new ObjectName(ManagementFactory.OPERATING_SYSTEM_MXBEAN_NAME);
		OperatingSystemMXBean osmxbean = ManagementFactory
				.getOperatingSystemMXBean();
		server.registerMBean(osmxbean, oname);
		// GarbageCollector Monitoring modules
		List list = ManagementFactory.getGarbageCollectorMXBeans();
		Iterator it = list.iterator();
		while (it.hasNext()) {
			GarbageCollectorMXBean gbmxbean = (GarbageCollectorMXBean) it
					.next();
			oname = new ObjectName(
					ManagementFactory.GARBAGE_COLLECTOR_MXBEAN_DOMAIN_TYPE
							+ ",name=" + gbmxbean.getName());
			server.registerMBean(gbmxbean, oname);
		}
		// MemoryManager Monitoring modules
		list = ManagementFactory.getMemoryManagerMXBeans();
		it = list.iterator();
		while (it.hasNext()) {
			MemoryManagerMXBean mmmxbean = (MemoryManagerMXBean) it.next();
			oname = new ObjectName(
					ManagementFactory.MEMORY_MANAGER_MXBEAN_DOMAIN_TYPE
							+ ",name=" + mmmxbean.getName());
			server.registerMBean(mmmxbean, oname);
		}
		// MemoryPool Monitoring modules
		list = ManagementFactory.getMemoryPoolMXBeans();
		it = list.iterator();
		while (it.hasNext()) {
			MemoryPoolMXBean mpmxbean = (MemoryPoolMXBean) it.next();
			oname = new ObjectName(
					ManagementFactory.MEMORY_POOL_MXBEAN_DOMAIN_TYPE + ",name="
							+ mpmxbean.getName());
			server.registerMBean(mpmxbean, oname);
		}
		// Loggin Monitoring module
		LoggingMXBean lmxbean = LogManager.getLoggingMXBean();
		oname = new ObjectName(LogManager.LOGGING_MXBEAN_NAME);
		server.registerMBean(lmxbean, oname);
	}

	/**
	 * Create a new RMI connector
	 * 
	 * @param listenURL
	 *            String of url
	 */
	public void createRMIConnector(String listenURL) {
		try {
			JMXServiceURL url = new JMXServiceURL(
					"service:jmx:rmi:///jndi/rmi://" + listenURL);
			cs = JMXConnectorServerFactory.newJMXConnectorServer(url, null,
					server);
		}
		catch (MalformedURLException e) {
			e.printStackTrace();
			System.exit(0);
		}
		catch (IOException e) {
			log("IOException: " + e.getMessage());
		}
	}

	/**
	 * Starts the RMI connector. Makes it listen for connections
	 * 
	 * @return true or false
	 */
	public Boolean startRMIServer() {
		try {
			cs.start();
			return true;
		}
		catch (IOException e) {
			log("Error: The RMI service could not be started. Is the rmiregistry running?");
		}
		return false;
	}

	/**
	 * Stops the RMI connector from listening for connections
	 */
	public void stopRMIServer() {
		try {
			cs.stop();
			log("RMI service has been stopped");
		}
		catch (IOException e) {
			e.printStackTrace();
			System.exit(0);
		}
	}
}
