package coms.mbeans;

import java.io.IOException;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import coms.core.Config;
import coms.database.ConnectionException;
import coms.database.Database;
import coms.metadata.Table;
import coms.metadata.TableField;

/**
 * Gives out access to the oracle database in the form of OracleConnections.
 * Will also monitor the Oracle database
 * 
 * @author Tom Arild Jakobsen & Mats Mortensen
 */
public class Oracle extends Database implements OracleMBean {
	private final String configUrl = "./config/database/OracleConfig.xml";
	private final String driver = "oracle.jdbc.driver.OracleDriver";
	private String protocol = "jdbc:oracle:thin:@";
	private Properties props;

	/**
	 * Creates an Derby object
	 */
	public Oracle() {
		super();
		try {
			// Parses the configuration file
			// TODO endre hvor pathen begynner i ant fila?
			HashMap<String, String> config = Config.read(configUrl);
			// Create connection properties
			props = new Properties();
			props.put("user", config.get("user"));
			props.put("password", config.get("pass"));
			protocol += config.get("host") + ":" + config.get("port") + ":"
					+ config.get("name");
			// Initiates database
			init(driver, protocol, props);
		}
		catch (IOException e) {
			// Should not occur
			e.printStackTrace();
		}
		catch (Exception e) {
			// Should not occur
			e.printStackTrace();
		}
	}

	/**
	 * Shut downs the database
	 */
	public void shutdown() throws SQLException, ConnectionException {
		shutdownDB(null);
	}

	/**
	 * Uses the database to create a Table object to the corresponding name
	 * 
	 * @param name
	 *            the name of the table to get info about
	 * @return the complete metainfo of the table
	 */
	public Table getTableInfo(String table) throws SQLException,
			ConnectionException {
		// Selects the column names from a table in the database
		String sqlStr = "SELECT column_name, data_type, data_default FROM all_tab_columns WHERE table_name='"
				+ table.toUpperCase() + "'";
		return getTableDesc(table, sqlStr, "column_name", "data_type",
				"data_default");
	}

	/**
	 * Adds a table to the database
	 * 
	 * @param tbl
	 *            The table to be added to the database
	 * @return wheater the table was added successfully
	 */
	protected String getCreateTableSql(Table tbl) throws SQLException,
			ConnectionException {
		String def;
		TableField primary = null;
		primary = tbl.getPrimaryKey();
		// Create Auto Increment sequence
		// http://www.techonthenet.com/oracle/sequences.php
		/*
		 * if(tbl.hasField("ID")){ try{ sqlStr = "CREATE SEQUENCE
		 * "+tbl.getName().toUpperCase()+"_SEQ MINVALUE 1 START WITH 1 INCREMENT
		 * BY 1 CACHE 20"; con.executeSQL(sqlStr, false); }catch(SQLException
		 * e){ // Already exists, drop and recreate sqlStr = "DROP SEQUENCE
		 * "+tbl.getName().toUpperCase()+"_SEQ"; con.executeSQL(sqlStr, false);
		 * sqlStr = "CREATE SEQUENCE "+tbl.getName().toUpperCase()+"_SEQ
		 * MINVALUE 1 START WITH 1 INCREMENT BY 1 CACHE 20";
		 * con.executeSQL(sqlStr, false); } }
		 */
		String sqlStr = "CREATE TABLE " + tbl.getName() + "(";
		// Add primary key
		sqlStr += primary.getName() + " " + toSqlType(primary.getType());
		sqlStr += " NOT NULL, ";
		// Iterates over the fields in the table
		Iterator it = tbl.getFieldIterator();
		while (it.hasNext()) {
			TableField field = (TableField) it.next();
			// Adds all the non primary fields to the sql string
			if (!field.isPrimaryKey()) {
				sqlStr += field.getName() + " " + toSqlType(field.getType());
				// If there is a default value
				def = field.getDefault();
				if (def != null && !def.equals(""))
					sqlStr += " DEFAULT " + def;
				sqlStr += ", ";
			}
		}
		sqlStr += " PRIMARY KEY(" + primary.getName() + ")";
		sqlStr += ")";
		return sqlStr;
	}

	/**
	 * Dropps a table from the database
	 * 
	 * @return wheater the table was dropped successfully
	 */
	public Boolean dropTable(Table tbl) throws SQLException,
			ConnectionException {
		String sqlStr = "DROP TABLE " + tbl.getName();
		return dropDBTable(sqlStr);
	}

	/**
	 * Converts a java type to a sql type
	 * 
	 * @param type
	 *            the java type to be converted
	 * @return the sql equivalent to the java type
	 */
	protected String toSqlType(String type) {
		if (type.equalsIgnoreCase("string"))
			return "VARCHAR(200)";
		else if (type.equalsIgnoreCase("bigdecimal"))
			return "DECIMAL(0, 15)";
		else if (type.equalsIgnoreCase("boolean"))
			return "BIT";
		else if (type.equalsIgnoreCase("byte"))
			return "CHAR";
		else if (type.equalsIgnoreCase("bytes"))
			return "LONG RAW";
		else if (type.equalsIgnoreCase("int"))
			return "NUMBER";
		else if (type.equalsIgnoreCase("integer"))
			return "NUMBER";
		else if (type.equalsIgnoreCase("short"))
			return "NUMBER";
		else if (type.equalsIgnoreCase("long"))
			return "LONG";
		else if (type.equalsIgnoreCase("double"))
			return "FLOAT";
		else if (type.equalsIgnoreCase("float"))
			return "FLOAT";
		else if (type.equalsIgnoreCase("date"))
			return "DATE";
		else if (type.equalsIgnoreCase("time"))
			return "TIME";
		else if (type.equalsIgnoreCase("timestamp"))
			return "TIMESTAMP";
		else
			return type;
	}
}
