package coms.mbeans;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import javax.management.ObjectName;
import coms.component.RequestResponse;
import coms.core.DeploymentException;
import coms.core.Module;
import coms.metadata.MetaData;
import coms.metadata.Service;
import coms.metadata.ServiceDescription;
import coms.metadata.ServiceOperation;

public class ServiceDescriptor extends Module implements ServiceDescriptorMBean {
	private HashMap<String, ServiceDescription> serviceDescMap;
	private HashMap<String, Service> serviceMap;

	public ServiceDescriptor() {
		serviceDescMap = new HashMap<String, ServiceDescription>();
		serviceMap = new HashMap<String, Service>();
	}

	public String addServiceDescription(ServiceDescription serviceDesc,
			HashMap<String, MetaData> metaMap) {
		try {
			// Add the service in the service description to the serviceMap and
			// verify the service
			Service service = serviceDesc.getService();
			// Verify the service
			verifyService(service, metaMap);
			// Check that the service name is unique
			if (!serviceMap.containsKey(service.getName()))
				serviceMap.put(service.getName(), service);
			else
				throw new DeploymentException(
						"All service names must be unique. The service name: "
								+ service.getName() + " is already in use");
			serviceDescMap.put(serviceDesc.getJarPackage(), serviceDesc);
			// Build dependency graph
			HashMap dependencyMap = serviceDesc.getDependencyMap();
			Iterator it = dependencyMap.keySet().iterator();
			// For each service
			while (it.hasNext()) {
				String key = (String) it.next();
				Service serv = serviceMap.get(key);
				List list = (List) dependencyMap.get(key);
				Iterator dit = list.iterator();
				// For each service it depends on
				while (dit.hasNext()) {
					String dkey = (String) dit.next();
					Service dservice = serviceMap.get(dkey);
					// Add it to the dependency list
					serv.addIDependOn(dservice);
				}
			}
			return "";
		}
		catch (DeploymentException e) {
			return e.toString();
		}
	}

	private void verifyService(Service service,
			HashMap<String, MetaData> metaMap) throws DeploymentException {
		MetaData component = null;
		ServiceOperation operation = null;
		try {
			HashMap map = service.getOperationMap();
			Set list = map.keySet();
			// Verify the methods (if they exist)
			Iterator it = list.iterator();
			while (it.hasNext()) {
				operation = (ServiceOperation) map.get((String) it.next());
				// Get the component used to process the user input
				component = (MetaData) metaMap
						.get(operation.getComponentName());
				// Check that it has declared the appropriate method in the
				// class
				Class componentClass = component.getComponentClass();
				Method method = componentClass.getDeclaredMethod(operation
						.getMethod(), Class.forName(HashMap.class.getName()));
				Class returnType = method.getReturnType();
				if (!returnType.equals(Class.forName(RequestResponse.class
						.getName())))
					throw new NoSuchMethodException(
							"Return type is not of type: RequestResponse");
				// Check that it has declared the appropriate method in its
				// mbean interface
				Class[] interfaces = componentClass.getInterfaces();
				Class mbeanInterface = null;
				for (int i = 0; i < interfaces.length; i++)
					if (interfaces[i].getName().equals(
							componentClass.getName() + "MBean"))
						mbeanInterface = interfaces[i];
				method = mbeanInterface.getDeclaredMethod(
						operation.getMethod(), Class.forName(HashMap.class
								.getName()));
				returnType = method.getReturnType();
				if (!returnType.equals(Class.forName(RequestResponse.class
						.getName())))
					throw new NoSuchMethodException();
			}
			// Verify that the components it uses are loaded
			List clist = service.getComponentList();
			it = clist.iterator();
			while (it.hasNext()) {
				String name = (String) it.next();
				MetaData com = (MetaData) metaMap.get(name);
				if (com == null)
					throw new DeploymentException("The service "
							+ service.getName() + " uses the component \""
							+ name + "\" which is not installed/loaded");
			}
		}
		catch (ClassNotFoundException e) {
			throw new DeploymentException(e.toString());
		}
		catch (NoSuchMethodException e) {
			e.printStackTrace();
			String msg = "The service " + service.getName()
					+ " uses the method " + operation.getMethod()
					+ " in component " + component.getName();
			msg += ", but this method is not declared (properly?) in component "
					+ component.getName()
					+ " or in its interface "
					+ component.getClassName() + "MBean";
			msg += "<br>Caused by: " + e.toString();
			throw new DeploymentException(msg);
		}
	}

	public HashMap<String, ServiceDescription> getServiceDescriptions() {
		return serviceDescMap;
	}

	public void enableService(String pkg, String serviceName) {
		Service service;
		if (serviceMap.containsKey(serviceName))
			service = serviceMap.get(serviceName);
		else
			return;
		// Get all services this service is dependent on
		List list = service.getIDependOn();
		Iterator it = list.iterator();
		// Enable all services this service is dependent on
		while (it.hasNext()) {
			Service dep = (Service) it.next();
			enableService(pkg, dep.getName());
		}
		// Get the components this service uses
		list = service.getComponentList();
		it = list.iterator();
		// Start all components
		while (it.hasNext())
			invokeComponent((String) it.next(), "startComponent");
		// Set service to enabled
		service.setEnabled(true);
	}

	private void invokeComponent(String component, String method) {
		try {
			ObjectName manager = createObjectName("ComponentManager", "Core");
			server.invoke(manager, method, new Object[] { component },
					new String[] { String.class.getName() });
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}

	public void disableService(String pkg, String serviceName) {
		Service service = null;
		if (serviceMap.containsKey(serviceName))
			service = serviceMap.get(serviceName);
		else
			return;
		// Get services that depend on this service
		List list = service.getDependOnMe();
		Iterator it = list.iterator();
		// Disable all services that this service depend on
		while (it.hasNext()) {
			Service dep = (Service) it.next();
			disableService(pkg, dep.getName());
		}
		// Get the components this service uses
		list = service.getComponentList();
		it = list.iterator();
		// Stop all components
		while (it.hasNext())
			invokeComponent((String) it.next(), "stopComponent");
		// Set the service to disabled
		service.setEnabled(false);
	}
}
