package coms.metadata;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;
import coms.core.JarLoader;
import coms.core.JarLoaderException;
import coms.core.JarValidityException;

public class DeploymentDescriptorParser implements ContentHandler {
	private XMLReader parser;
	private MetaData data;
	private String starttag, endtag;
	private String fullvalue = "";
	private String state1 = "";
	private String state2 = "";
	private DeploymentDescription deployDesc;
	private LinkedHashMap<String, MetaData> deployList = new LinkedHashMap<String, MetaData>();
	private ServiceDescription serviceDesc;
	private Service service;
	private ServiceOperation operation;
	private String pkg = "";
	private HashMap<String, ArrayList<String>> dependencyMap = new HashMap<String, ArrayList<String>>();

	// Takes appropriate action when we have a starttag of some type
	public void startElement(String namespaceURI, String localName,
			String qName, Attributes atts) throws SAXException {
		starttag = localName.toLowerCase();
		if (starttag.equals("service")) {
			state1 = starttag;
			serviceDesc = new ServiceDescription(pkg);
			service = new Service();
		}
		else if (starttag.equals("operation")) {
			state2 = starttag;
			operation = new ServiceOperation();
		}
		else if (starttag.equals("depends_on")) {
			state2 = starttag;
		}
		else if (starttag.equals("deploy")) {
			state1 = starttag;
		}
		else if (starttag.equals("component")) {
			state2 = starttag;
		}
		if (starttag.equals("component") && state1.equals("deploy")) {
			data = new MetaData();
			data.setFileName(pkg);
		}
	}

	// Takes appropriate action when we have a endtag of come type
	public void endElement(String namespaceURI, String localName, String qName)
			throws SAXException {
		endtag = localName.toLowerCase();
		starttag = "";
		if (endtag.equals("name") && state1.equals("service"))
			service.setName(fullvalue);
		if (endtag.equals("version") && state1.equals("service"))
			service.setVersion(fullvalue);
		if (endtag.equals("description") && state1.equals("service"))
			service.setDescription(fullvalue);
		if (endtag.equals("dbtype") && state1.equals("service"))
			service.setDbType(fullvalue);
		if (endtag.equals("display_name") && state2.equals("operation")
				&& state1.equals("service"))
			operation.setName(fullvalue);
		if (endtag.equals("cname") && state2.equals("operation")
				&& state1.equals("service"))
			operation.setComponentName(fullvalue);
		if (endtag.equals("ctype") && state2.equals("operation")
				&& state1.equals("service"))
			operation.setComponentType(fullvalue);
		if (endtag.equals("method") && state2.equals("operation")
				&& state1.equals("service"))
			operation.setMethod(fullvalue);
		if (endtag.equals("operation"))
			service.addOperation(operation);
		if (endtag.equals("dservice") && state2.equals("depends_on")
				&& state1.equals("service"))
			buildDependencyMap();
		if (endtag.equals("service")) {
			serviceDesc.setService(service);
			serviceDesc.setDependencyMap(dependencyMap);
			deployDesc.setServiceDescription(serviceDesc);
		}
		if (endtag.equals("name") && state2.equals("component")
				&& state1.equals("deploy"))
			data.setName(fullvalue);
		if (endtag.equals("class") && state2.equals("component")
				&& state1.equals("deploy"))
			data.setClassName(fullvalue);
		if (endtag.equals("metafile") && state2.equals("component")
				&& state1.equals("deploy"))
			data.setMetaFileName(fullvalue);
		if (endtag.equals("component") && state1.equals("deploy")) {
			if (!deployList.containsKey(data.getName())) {
				deployList.put(data.getName(), data);
			}
			else
				throw new SAXException(
						"Two (or more) components have the same name: \""
								+ data.getName() + "\"");
		}
		if (endtag.equals("deploy"))
			deployDesc.setComponentMap(deployList);
		starttag = "";
		fullvalue = "";
		endtag = "";
	}

	// Makes a hashmap where the services are keys and their dependencies are
	// list values
	private void buildDependencyMap() {
		if (!dependencyMap.containsKey(service.getName())) {
			ArrayList<String> newlist = new ArrayList<String>();
			newlist.add(fullvalue);
			dependencyMap.put(service.getName(), newlist);
		}
		else {
			ArrayList<String> list = dependencyMap.get(service.getName());
			list.add(fullvalue);
			dependencyMap.put(service.getName(), list);
		}
	}

	// Gets the content between the start and end tag
	public void characters(char[] text, int start, int length)
			throws SAXException {
		String value = new String(text, start, length);
		value = value.trim();
		fullvalue += value;
	}

	public DeploymentDescriptorParser() throws SAXException {
		try {
			parser = XMLReaderFactory.createXMLReader();
			parser.setFeature("http://xml.org/sax/features/validation", true);
		}
		catch (SAXNotRecognizedException ex) {
			System.out.println("Installed XML parser cannot validate");
			System.exit(0);
		}
		catch (SAXNotSupportedException e) {
			System.out.println("Installed XML parser cannot validate");
			System.exit(0);
		}
		// Install the content handler
		parser.setContentHandler(this);
	}

	// Initiate parsing of xml document
	public DeploymentDescription parse(JarLoader jl, String xmlfile)
			throws SAXParseException, SAXException, JarLoaderException,
			JarValidityException {
		try {
			JarLoader jarLoader = jl;
			pkg = jarLoader.getFileName();
			// Set error handler
			parser.setErrorHandler(new ValidityErrorHandler(xmlfile));
			// Read the deployment descriptor
			InputStream stream = jarLoader.readXMLFile(xmlfile);
			// New DeploymentDescription object
			deployDesc = new DeploymentDescription(pkg);
			// Parse the xml file
			parser.parse(new InputSource(stream));
			return deployDesc;
		}
		catch (IOException e) {
			throw new JarLoaderException("Error reading xml data from: " + pkg
					+ " when parsing. \nCaused by: " + e.getMessage());
		}
	}

	// Initiate parsing of xml document
	public DeploymentDescription parse(File xmlfile, String jarpackage)
			throws SAXParseException, SAXException, JarLoaderException,
			JarValidityException {
		try {
			pkg = jarpackage;
			// Set error handler
			parser.setErrorHandler(new ValidityErrorHandler(xmlfile.getName()));
			// Read the deployment descriptor
			InputStream stream = new FileInputStream(xmlfile);
			// New DeploymentDescription object
			deployDesc = new DeploymentDescription(jarpackage);
			// Parse the xml file
			parser.parse(new InputSource(stream));
			return deployDesc;
		}
		catch (IOException e) {
			throw new JarLoaderException("Error reading xml data from: " + pkg
					+ " when parsing. \nCaused by: " + e.getMessage());
		}
	}

	public void setDocumentLocator(Locator locator) {}

	public void startDocument() throws SAXException {}

	public void endDocument() throws SAXException {}

	public void ignorableWhitespace(char[] text, int start, int length)
			throws SAXException {}

	public void processingInstruction(String target, String data)
			throws SAXException {}

	public void startPrefixMapping(String prefix, String uri)
			throws SAXException {}

	public void endPrefixMapping(String prefix) throws SAXException {}

	public void skippedEntity(String name) throws SAXException {}
}
