package coms.metadata;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;
import coms.core.JarLoader;
import coms.core.JarLoaderException;
import coms.core.JarValidityException;

public class MetaDataParser implements ContentHandler {
	private XMLReader parser;
	private JarLoader jarLoader;
	private MetaData data;
	private String starttag, endtag;
	private Table table;
	private TableField field;
	private HashMap<String, String> configmap;
	private Bind bind;
	private String state = "";
	private String mapkey;
	private String mapval;
	private String primaryKey = "";

	private StringBuffer tempValue = new StringBuffer();
	
	public void startElement(String namespaceURI, String localName,
			String qName, Attributes atts) throws SAXException {
		tempValue = new StringBuffer();
		
		starttag = localName.toLowerCase();
		if (starttag.equals("table")) {
			table = new Table();
			primaryKey = "";
		}
		else if (starttag.equals("field"))
			field = new TableField();
		else if (starttag.equals("config")) {
			state = "config";
			configmap = new HashMap<String, String>();
		}
		else if (starttag.equals("option")) {
			mapkey = "";
			mapval = "";
		}
		else if (starttag.equals("subscribe")) {
			state = "subscribe";
		}
		else if (starttag.equals("bind")) {
			bind = new Bind(atts.getValue("comp"), atts.getValue("to"), atts
					.getValue("class"));
			data.addBinding(bind);
		}
	}

	public void endElement(String namespaceURI, String localName, String qName)
			throws SAXException {
		
		/* XXX: see comment in charachter() */
		
		String value = tempValue.toString().trim();
		if (starttag.equals("version"))
			data.setVersion(value);
		else if (starttag.equals("type"))
			data.setType(value);
		else if (starttag.equals("description"))
			data.setDescription(value);
		else if (starttag.equals("fieldname"))
			field.setName(value);
		else if (starttag.equals("fieldtype"))
			field.setType(value);
		else if (starttag.equals("default"))
			field.setDefault(value);
		else if (starttag.equals("dname"))
			data.addDependency(value);
		else if (starttag.equals("tablename"))
			table.setName(value);
		else if (starttag.equals("primarykey"))
			primaryKey = value;
		else if (starttag.equals("bindingtype"))
			data.setBindingType(value);
		else if (state.equals("config")) {
			if (value.length() > 0) {
				if (starttag.equals("optionname"))
					mapkey = value;
				if (starttag.equals("optionvalue"))
					mapval = value;
			}
		}
		else if (state.equals("subscribe")) {
			if (value.length() > 0)
				if (starttag.equals("to_table"))
					data.addSubscription(value);
		}

		endtag = localName.toLowerCase();
		starttag = "";
		if (endtag.equals("field")) {
			if (field.getName().equalsIgnoreCase(primaryKey)) {
				table.addPrimaryKey(field.getName(), field.getType());
				primaryKey = "";
			}
			else
				table.addField(field);
		}
		else if (endtag.equals("table"))
			data.addTable(table);
		else if (endtag.equals("config"))
			state = "";
		else if (endtag.equals("subscribe"))
			state = "";
		else if (endtag.equals("option"))
			configmap.put(mapkey, mapval);
		endtag = "";
	}

	public void characters(char[] text, int start, int length)
			throws SAXException {
//		String value = new String(text, start, length);
//		value = value.trim();
		
		/** XXX: comment about sax parsing
		 * When saxparsing, and the characters event is called,
		 * there's no guarantee that all the characters in
		 * the text element is available at the same time.
		 * 
		 * For instance, when parsing the text element 
		 * 
		 * <fieldtype>TIMESTAMP</fieldtype>
		 * 
		 * it's just as likely to get one characters-event with TIM and
		 * one with ESTAMP as to get just one event with TIMESTAMP. 
		 * Therefore it makes sense to use whatever characters in a 
		 * text element when the endElement-event is reached, 
		 * since then all the characters have been read.
		 * 
		 */
		tempValue.append(new String(text, start, length));
		
	}

	public MetaDataParser(MetaData mdata, JarLoader jarLoader)
			throws SAXException {
		try {
			this.data = mdata;
			this.jarLoader = jarLoader;
			parser = XMLReaderFactory.createXMLReader();
			parser.setFeature("http://xml.org/sax/features/validation", true);
		}
		catch (SAXNotRecognizedException ex) {
			System.out.println("Installed XML parser cannot validate");
			System.exit(0);
		}
		catch (SAXNotSupportedException e) {
			System.out.println("Installed XML parser cannot validate");
			System.exit(0);
		}
		// Install the content handler
		parser.setContentHandler(this);
	}

	public MetaData parse(String xmlfile) throws SAXParseException,
			SAXException, JarLoaderException, JarValidityException {
		try {
			// Install a new error handler
			parser.setErrorHandler(new ValidityErrorHandler(xmlfile));
			InputStream stream = jarLoader.readXMLFile(xmlfile);
			parser.parse(new InputSource(stream));
			data.setCustomOptions(configmap);
			return data;
		}
		catch (IOException e) {
			throw new JarLoaderException("Error reading the xml from: "
					+ xmlfile + " when parsing. \nCaused by: " + e.getMessage());
		}
	}

	public MetaData parseFile(File xmlfile) throws SAXParseException,
			SAXException, JarLoaderException, JarValidityException {
		try {
			// Install a new error handler
			parser.setErrorHandler(new ValidityErrorHandler(xmlfile.getName()));
			InputStream stream = new FileInputStream(xmlfile);
			parser.parse(new InputSource(stream));
			data.setCustomOptions(configmap);
			return data;
		}
		catch (IOException e) {
			throw new JarLoaderException("Error reading the xml from: "
					+ xmlfile + " when parsing. \nCaused by: " + e.getMessage());
		}
	}

	public void setDocumentLocator(Locator locator) {}

	public void startDocument() throws SAXException {}

	public void endDocument() throws SAXException {}

	public void ignorableWhitespace(char[] text, int start, int length)
			throws SAXException {}

	public void processingInstruction(String target, String data)
			throws SAXException {}

	public void startPrefixMapping(String prefix, String uri)
			throws SAXException {}

	public void endPrefixMapping(String prefix) throws SAXException {}

	public void skippedEntity(String name) throws SAXException {}
}
