package coms.metadata;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * An object to keep a table for a relational database
 * 
 * @author Tom Arild Jakobsen & Mats Mortensen
 */
public class Table {
	private String name;
	private Set<TableField> fields;

	/**
	 * Contructor for a table object. Creates a table without a name
	 */
	public Table() {
		this("");
	}

	/**
	 * Constructor for a Table object
	 * 
	 * @param n
	 *            the name of the table
	 */
	public Table(String n) {
		name = n.toUpperCase();
		fields = new HashSet<TableField>();
	}

	/**
	 * Compares wheter the name and fields of a table match this one
	 * 
	 * @param tbl
	 *            the table to be checked against this one
	 * @return wheater the other table equals this or not
	 */
	public boolean equals(Table tbl) {
		boolean isEqual = true;
		int size1 = 0, size2 = 0;
		size1 = tbl.length();
		size2 = length();
		// Ignore potentially auto created ID field
		if (tbl.hasField("ID"))
			size1--;
		if (hasField("ID"))
			size2--;
		if (!name.equalsIgnoreCase(tbl.getName()))
			isEqual = false;
		if (size1 != size2)
			isEqual = false;
		Iterator it = tbl.getFieldIterator();
		while (it.hasNext()) {
			TableField tf = (TableField) it.next();
			if (!tf.getName().equalsIgnoreCase("ID"))
				if (!hasField(tf.getName()))
					isEqual = false;
		}
		return isEqual;
	}

	/**
	 * Adds another field to the table
	 * 
	 * @param n
	 *            name of the field
	 * @param t
	 *            type of the field
	 * @param d
	 *            default value of the field
	 */
	public void addField(String n, String t, String d) {
		addField(new TableField(n, t, d));
	}

	/**
	 * Adds another field to the table
	 * 
	 * @param f
	 *            Field to be added to the table
	 */
	public void addField(TableField f) {
		fields.add(f);
	}

	/**
	 * Adds another field to the table
	 * 
	 * @param n
	 *            name of the field
	 * @param t
	 *            type of the field
	 */
	public void addPrimaryKey(String n, String t) {
		addField(new TableField(n, t, null, true));
	}

	/**
	 * Gets the name of the primary key in the table
	 * 
	 * @return the name of the primary key
	 */
	public TableField getPrimaryKey() {
		TableField primKey = null;
		Iterator it = fields.iterator();
		while (it.hasNext()) {
			TableField tf = (TableField) it.next();
			if (tf.isPrimaryKey())
				primKey = tf;
		}
		return primKey;
	}

	public boolean hasPrimaryKey() {
		Iterator it = getFieldIterator();
		while (it.hasNext()) {
			TableField field = (TableField) it.next();
			if (field.isPrimaryKey())
				return true;
		}
		return false;
	}

	/**
	 * Checks if the table has a field with a give name
	 * 
	 * @param fname
	 *            the name to check if the table has
	 * @return wheater the table has the field
	 */
	public boolean hasField(String fname) {
		Iterator it = fields.iterator();
		while (it.hasNext()) {
			TableField tf = (TableField) it.next();
			if (tf.getName().toUpperCase().equals(fname.toUpperCase()))
				return true;
		}
		return false;
	}

	/**
	 * Returns a specific field in the table
	 * 
	 * @param fname
	 *            the name of the field to be returned
	 * @return a specific field in the table
	 */
	public TableField getField(String fname) {
		fname = fname.toUpperCase();
		Iterator it = fields.iterator();
		TableField tf;
		while (it.hasNext()) {
			tf = (TableField) it.next();
			if (fname.equals(tf.getName())) {
				return tf;
			}
		}
		return null;
	}

	/**
	 * Gives access to the number of fields
	 * 
	 * @return the number of fields in the table
	 */
	public int length() {
		return fields.size();
	}

	/**
	 * Gives access to the name of the table
	 * 
	 * @return the name of the table
	 */
	public String getName() {
		return name;
	}

	/**
	 * Gives access to setting a new table name
	 * 
	 * @param n
	 *            the new table name
	 */
	public void setName(String n) {
		name = n;
	}

	/**
	 * Gives access to a iterator over the fields
	 * 
	 * @return a iterator over the fields
	 */
	public Iterator getFieldIterator() {
		return this.fields.iterator();
	}
}
