package coms.tools.relational;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import coms.CoMS;
import coms.database.ConnectionException;
import coms.database.DatabaseConnection;

/**
 * Delete is used against relational databases to delete a piece of data.
 * 
 * @author Tom Arild Jakobsen & Mats Mortensen
 */
public class Delete {
	public static final int EQUAL = 0;
	public static final int NOT = 1;
	public static final int SMALLER = 2;
	public static final int GREATER = 3;
	public static final int LIKE = 4;
	private ArrayList<Object> params = new ArrayList<Object>();
	private ArrayList<String> cond = new ArrayList<String>();
	private String tbl;
	private String dbType;

	public Delete(String tableName, String DBType) {
		this.tbl = tableName;
		this.dbType = DBType;
	}

	/**
	 * Adds another value to the sql string
	 * 
	 * @param field
	 *            the field to be compared against a value
	 * @param condition
	 *            the compare condition. Can be either: Constant.equal,
	 *            Constant.not, Constant.greater, Constant.smaller
	 * @param value
	 *            the value field is compared to
	 */
	public void and(String field, int condition, Object value) {
		setCondition(field, condition, value, " AND");
	}

	/**
	 * Puts this condition as a or statement
	 * 
	 * @param field
	 *            the field to be compared against a value
	 * @param condition
	 *            the compare condition. Can be either: Constant.equal,
	 *            Constant.not, Constant.greater, Constant.smaller
	 * @param value
	 *            the value field is compared to
	 */
	public void or(String field, int condition, Object value) {
		setCondition(field, condition, value, " OR");
	}

	/**
	 * Will not delete when this condition is true
	 * 
	 * @param field
	 *            the field to be compared against a value
	 * @param condition
	 *            the compare condition. Can be either: Constant.equal,
	 *            Constant.not, Constant.greater, Constant.smaller
	 * @param value
	 *            the value field is compared to
	 */
	public void andNot(String field, int condition, Object value) {
		setCondition(field, condition, value, " AND NOT");
	}

	/**
	 * Will not delete when this condition is true
	 * 
	 * @param field
	 *            the field to be compared against a value
	 * @param condition
	 *            the compare condition. Can be either: Constant.equal,
	 *            Constant.not, Constant.greater, Constant.smaller
	 * @param value
	 *            the value field is compared to
	 */
	public void orNot(String field, int condition, Object value) {
		setCondition(field, condition, value, " OR NOT");
	}

	/**
	 * Undoes the last condition set
	 */
	public void undo() {
		cond.remove(cond.size() - 1);
		params.remove(params.size() - 1);
	}

	/**
	 * Will execute the options put into Delete against a relational database.
	 * 
	 * @param tbl
	 *            The table on which the select will be executed
	 * @param dbType
	 *            The type of the database, can be found in
	 *            "getMetaData().getDbType()"
	 * @throws SQLException
	 *             if there is something wrong in the generated SQL string
	 * @throws ConnectionException
	 *             if a connection cannot be aquired
	 */
	public void execute() throws SQLException, ConnectionException {
		// Combines the different elements to a sql string
		String sqlStr = "DELETE FROM " + tbl;
		// If there are conditions, add them to the string
		if (!cond.isEmpty()) {
			sqlStr += " WHERE";
			Iterator it = cond.iterator();
			sqlStr += (String) it.next();
			while (it.hasNext())
				sqlStr += (String) it.next();
		}
		try {
			// Executes the sql string on a database connection
			DatabaseConnection con = getDatabaseConnection(dbType);
			con.executeSQL(sqlStr, params.toArray(), false);
			con.release();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	private DatabaseConnection getDatabaseConnection(String type)
			throws Exception {
		MBeanServer server = CoMS.getServerHandle();
		ObjectName db = new ObjectName(server.getDefaultDomain() + ":Name="
				+ type + ", Type=Core");
		return (DatabaseConnection) server.getAttribute(db, "Connection");
	}

	/**
	 * Creates a substring containing the information in this condition
	 * 
	 * @param field
	 *            in the database to be quaried against
	 * @param condition
	 *            which boolean expression to be used in the checking
	 * @param value
	 *            to be checked against
	 * @param type
	 *            can be either AND, OR or NOT
	 */
	private void setCondition(String field, int condition, Object value,
			String type) {
		String sqlStr = "";
		// If this is the first condition, don't use prefix AND or OR
		if (cond.size() == 0) {
			// If the first condition is a NOT, drop AND and OR, but keep NOT
			if (type.equals(" AND NOT") || type.equals(" OR NOT")) {
				sqlStr = " NOT";
			}
		}
		else
			sqlStr = type;
		sqlStr += " " + field;
		if (condition == GREATER)
			sqlStr += " > ?";
		else if (condition == SMALLER)
			sqlStr += " < ?";
		else if (condition == NOT)
			sqlStr += " != ?";
		else if (condition == LIKE)
			sqlStr += " LIKE ?";
		else
			sqlStr += " = ?";
		cond.add(sqlStr);
		params.add(value);
	}
}
