package coms.tools.relational;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import coms.CoMS;
import coms.database.ConnectionException;
import coms.database.DatabaseConnection;

/**
 * Insert is used to insert data into a table in a relational database.
 * 
 * @author Tom Arild Jakobsen & Mats Mortensen
 */
public class Insert {
	private String prevCol = null;
	private Object prevVal = null;
	private HashMap<String, Object> data = new HashMap<String, Object>();
	private String tbl;
	private String dbType;

	public Insert(String tableName, String DBType) {
		this.tbl = tableName;
		this.dbType = DBType;
	}

	/**
	 * Inserts the value into the column
	 * 
	 * @param col
	 *            column where the value is to be placed
	 * @param value
	 *            to be put into the database
	 */
	public void column(String col, Object value) {
		prevCol = col;
		prevVal = data.get(col);
		data.put(col, value);
	}

	/**
	 * Removes the last value to be put into the database
	 */
	public void undo() {
		if (prevVal == null)
			data.remove(prevCol);
		else
			data.put(prevCol, prevVal);
	}

	/**
	 * Will put the previously specified data into the database.
	 * 
	 * @param tbl
	 *            The table the data will be inserted into
	 * @param dbType
	 *            The type of the database, can be found in
	 *            "getMetaData().getDbType()
	 * @throws SQLException
	 *             if there is something wrong in the generated SQL string
	 * @throws ConnectionException
	 *             if a connection cannot be aquired
	 */
	public void execute() throws SQLException, ConnectionException {
		Iterator it = data.keySet().iterator();
		String valStr = "";
		String sqlStr = "INSERT INTO " + tbl + " (";
		// Iterates over the data picking out the fields in each row
		while (it.hasNext()) {
			// Drops the commas the first time
			if (!valStr.equals("")) {
				valStr += ", ";
				sqlStr += ", ";
			}
			sqlStr += it.next();
			valStr += "?";
		}
		sqlStr += ") VALUES (" + valStr + ")";
		try {
			DatabaseConnection con = getDatabaseConnection(dbType);
			con.executeSQL(sqlStr, data.values().toArray(), false);
			con.release();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		// Sends an event that there has been done changes to table
		notifyEvent(tbl);
	}

	/**
	 * Puts a table containing values into a database
	 * 
	 * @param tbl
	 *            The table to be put into the database
	 * @param dbType
	 *            The type of the database, can be found in
	 *            "getMetaData().getDbType()
	 * @param data
	 *            Component that called this method
	 * @throws SQLException
	 *             if there is something wrong in the generated SQL string
	 * @throws ConnectionException
	 *             if a connection cannot be aquired
	 */
	public static void execute(String tbl, String dbType,
			Set<Map<String, Object>> data) throws SQLException,
			ConnectionException {
		ArrayList<Object> val = new ArrayList<Object>();
		Map row;
		String fieldName;
		String valStr = "";
		String sqlStr = "INSERT INTO " + tbl + " (";
		Iterator field;
		Iterator it = data.iterator();
		// Iterates over the data picking out the fields in each row
		while (it.hasNext()) {
			row = (Map) it.next();
			field = row.keySet().iterator();
			while (field.hasNext()) {
				// Drops the commas the first time
				if (!valStr.equals("")) {
					valStr += ", ";
					sqlStr += ", ";
				}
				fieldName = (String) field.next();
				sqlStr += fieldName;
				valStr += "?";
				val.add(row.get(fieldName));
			}
		}
		sqlStr += ") VALUES (" + valStr + ")";
		try {
			DatabaseConnection con = getDatabaseConnection(dbType);
			con.executeSQL(sqlStr, val.toArray(), false);
			con.release();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		// Sends an event that there has been done changes to table
		notifyEvent(tbl);
	}

	private static DatabaseConnection getDatabaseConnection(String type)
			throws Exception {
		MBeanServer server = CoMS.getServerHandle();
		ObjectName db = new ObjectName(server.getDefaultDomain() + ":Name="
				+ type + ", Type=Core");
		return (DatabaseConnection) server.getAttribute(db, "Connection");
	}

	/**
	 * Notifies all components of a table change using the EventHandler.
	 * 
	 * @param String
	 *            table
	 */
	private static void notifyEvent(String table) throws ConnectionException {
		MBeanServer server = CoMS.getServerHandle();
		try {
			// If a server is found, choose which type of database we want to
			// contact
			if (server != null) {
				ObjectName componentAgent = new ObjectName(server
						.getDefaultDomain()
						+ ":Name=EventHandler, Type=Core");
				server.invoke(componentAgent, "notifyEvent",
						new Object[] { table }, new String[] { String.class
								.getName() });
			}
			else
				throw new ConnectionException("Could not find mbeanserver");
		}
		catch (Exception e) {
			// This should not occur
			e.printStackTrace();
		}
	}
}
