package coms.tools.relational;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import coms.CoMS;
import coms.database.ConnectionException;
import coms.database.DatabaseConnection;

/**
 * Select is used against relational databases to select a piece of data.
 * 
 * @author Tom Arild Jakobsen & Mats Mortensen
 */
public class Select {
	public static final int EQUAL = 0;
	public static final int NOT = 1;
	public static final int SMALLER = 2;
	public static final int GREATER = 3;
	public static final int LIKE = 4;
	public static final int MAX = 5;
	public static final int MIN = 6;
	public static final int ANY = 7;
	public static final int ASCENDING = 0;
	public static final int DESCENDING = 1;
	private String tbl;
	private String dbType;
	private String order = "";
	private ArrayList<String> column = new ArrayList<String>();
	private ArrayList<Object> params = new ArrayList<Object>();
	private ArrayList<String> cond = new ArrayList<String>();

	public Select(String tableName, String DBtype) {
		this.tbl = tableName;
		this.dbType = DBtype;
	}

	/**
	 * Sets a column in the table to be selected
	 * 
	 * @param col
	 *            Name of column to be selected
	 */
	public void column(String col) {
		column.add(col);
	}

	public void column(String col, int condition) {
		if (condition == MAX)
			column.add(" MAX(" + col + ")");
		else if (condition == MIN)
			column.add(" MIN(" + col + ")");
	}

	/**
	 * Adds another value to the sql string
	 * 
	 * @param field
	 *            the field to be compared against a value
	 * @param condition
	 *            the compare condition. Can be either: Constant.equal,
	 *            Constant.not, Constant.greater, Constant.smaller
	 * @param value
	 *            the value field is compared to
	 */
	public void and(String field, int condition, Object value) {
		setCondition(field, condition, value, " AND");
	}

	/**
	 * Puts this condition as a or statement
	 * 
	 * @param field
	 *            the field to be compared against a value
	 * @param condition
	 *            the compare condition. Can be either: Constant.equal,
	 *            Constant.not, Constant.greater, Constant.smaller
	 * @param value
	 *            the value field is compared to
	 */
	public void or(String field, int condition, Object value) {
		setCondition(field, condition, value, " OR");
	}

	/**
	 * Will not select when this condition is true
	 * 
	 * @param field
	 *            the field to be compared against a value
	 * @param condition
	 *            the compare condition. Can be either: Constant.equal,
	 *            Constant.not, Constant.greater, Constant.smaller
	 * @param value
	 *            the value field is compared to
	 */
	public void andNot(String field, int condition, Object value) {
		setCondition(field, condition, value, " AND NOT");
	}

	/**
	 * Will not select when this condition is true
	 * 
	 * @param field
	 *            the field to be compared against a value
	 * @param condition
	 *            the compare condition. Can be either: Constant.equal,
	 *            Constant.not, Constant.greater, Constant.smaller
	 * @param value
	 *            the value field is compared to
	 */
	public void orNot(String field, int condition, Object value) {
		setCondition(field, condition, value, " OR NOT");
	}

	/**
	 * Chooses which way the result will be ordered
	 * 
	 * @param col
	 *            the column the result will be ordered by
	 * @param dir
	 *            The direction of the result. Either Select.ASCENDING or
	 *            Select.DECENDING
	 */
	public void orderBy(String col, int dir) {
		if (dir == ASCENDING) {
			order = " ORDER BY " + col + " ASC";
		}
		else if (dir == DESCENDING) {
			order = " ORDER BY " + col + " DESC";
		}
	}

	/**
	 * Undoes the last condition set
	 */
	public void undo() {
		cond.remove(cond.size() - 1);
		params.remove(params.size() - 1);
	}

	/**
	 * Will execute the options put into select against a relational database.
	 * 
	 * @return the result from select done on the database
	 * @throws SQLException
	 *             if there is something wrong in the generated SQL string
	 * @throws ConnectionException
	 *             if a connection cannot be aquired
	 */
	public List<Map<String, Object>> execute() throws SQLException,
			ConnectionException {
		List<Map<String, Object>> result = null;
		Iterator it;
		String sqlStr;
		// Combines the different elements to a sql string
		sqlStr = "SELECT ";
		// If there no specific column is to be selected, select all
		if (column.isEmpty())
			sqlStr += "*";
		else {
			it = column.iterator();
			sqlStr += (String) it.next();
			while (it.hasNext())
				sqlStr += ", " + (String) it.next();
		}
		sqlStr += " FROM " + tbl;
		// If there are conditions, add them to the string
		if (!cond.isEmpty()) {
			sqlStr += " WHERE";
			it = cond.iterator();
			sqlStr += (String) it.next();
			while (it.hasNext())
				sqlStr += (String) it.next();
		}
		// Adds the chosen ordering if any
		sqlStr += order;
		try {
//			System.out.println(sqlStr);
			// Executes the sql string on a database connection
			DatabaseConnection con = getDatabaseConnection(dbType);
			result = con.executeSQL(sqlStr, params.toArray(), true);
			con.release();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		return result;
	}

	private DatabaseConnection getDatabaseConnection(String type)
			throws Exception {
		MBeanServer server = CoMS.getServerHandle();
		ObjectName db = new ObjectName(server.getDefaultDomain() + ":Name="
				+ type + ", Type=Core");
		return (DatabaseConnection) server.getAttribute(db, "Connection");
	}

	/**
	 * Creates a substring containing the information in this condition
	 * 
	 * @param field
	 *            in the database to be quaried against
	 * @param condition
	 *            which boolean expression to be used in the checking
	 * @param value
	 *            to be checked against
	 * @param type
	 *            can be either AND, OR or NOT
	 */
	private void setCondition(String field, int condition, Object value,
			String type) {
		String sqlStr = "";
		// If this is the first condition, don't use prefix AND or OR
		if (cond.size() == 0) {
			// If the first condition is a NOT, drop AND and OR, but keep NOT
			if (type.equals(" AND NOT") || type.equals(" OR NOT")) {
				sqlStr = " NOT";
			}
		}
		else
			sqlStr = type;
		sqlStr += " " + field;
		if (condition == GREATER)
			sqlStr += " > ?";
		else if (condition == SMALLER)
			sqlStr += " < ?";
		else if (condition == NOT)
			sqlStr += " != ?";
		else if (condition == LIKE)
			sqlStr += " LIKE ?";
		else
			sqlStr += " = ?";
		cond.add(sqlStr);
		params.add(value);
	}
}
